# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
Abstract transport model
++++++++++++++++++++++++

The transport layer submodule defines a high-level interface that abstracts transport-specific implementation
details from the transport-agnostic library core.
The main component is the interface class :class:`pyuavcan.transport.Transport`
accompanied by several auxiliary entities encapsulating and modeling different aspects of the
UAVCAN protocol stack, particularly:

- :class:`pyuavcan.transport.Session`
- :class:`pyuavcan.transport.Transfer`
- :class:`pyuavcan.transport.DataSpecifier`
- :class:`pyuavcan.transport.SessionSpecifier`
- :class:`pyuavcan.transport.PayloadMetadata`
- :class:`pyuavcan.transport.Priority`

These classes are specifically designed to map well onto the UAVCAN v1 transport layer model
(first discussed in this post: https://forum.uavcan.org/t/alternative-transport-protocols/324).
The following transfer metadata taxonomy table is the essence of the model;
one can map it onto aforementioned auxiliary definitions:

+----------------------+-------------------+-------------------+---------------------------------------+
|  Transfer            |                   |                   |                                       |
|  metadata taxonomy   |     Messages      |     Services      |              Comments                 |
+======================+===================+===================+=======================================+
|                      |          Transfer priority            | Not used above the transport layer.   |
+----------+-----------+-------------------+-------------------+---------------------------------------+
|          | Route     |                   |  Source node-ID   | Transport route information. If the   |
|          | specifier |  Source node-ID   +-------------------+ destination node-ID is not provided,  |
|          |           |                   |Destination node-ID| broadcast is implied.                 |
|Session   +-----------+-------------------+-------------------+---------------------------------------+
|specifier |           |                Kind                   | Contained information: kind of        |
|          | Data      +-------------------+-------------------+ transfer (message or service);        |
|          | specifier |                   |    Service-ID     | subject-ID for messages;              |
|          |           |    Subject-ID     +---------+---------+ service-ID with request/response      |
|          |           |                   | Request |Response | role selector for services.           |
+----------+-----------+-------------------+---------+---------+---------------------------------------+
|                      |             Transfer-ID               | Transfer sequence number.             |
+----------------------+---------------------------------------+---------------------------------------+


Sessions
++++++++

PyUAVCAN transport heavily relies on the concept of *session*.
In PyUAVCAN, session represents a **flow of data through the network defined by a particular
session specifier that either originates or terminates at the local node**.
Whenever the application desires to establish communication
(such as subscribing to a subject or invoking a service),
it commands the transport layer to open a particular session.
The session abstraction is sufficiently high-level to permit efficient mapping to features
natively available to concrete transport implementations.
For example, the UAVCAN/CAN transport uses the set of active input sessions to automatically compute the
optimal hardware acceptance filter configuration;
the UAVCAN/UDP transport can map sessions onto UDP port numbers,
establishing close equivalence between sessions and Berkeley sockets.

There can be at most one session per session specifier.
When a transport is requested to provide a session, it will first check if there is one for the specifier,
and return the existing one if so; otherwise, a new session will be created, stored, and returned.
Once created, the session will remain active until explicitly closed, or until the transport instance
that owns it is closed.

An output session that doesn't have a remote node-ID specified is called a *broadcast session*;
the opposite is called a *unicast session*.

An input session that doesn't have a remote node-ID specified is called a *promiscuous session*,
meaning that it accepts transfers with matching *data specifier* from any remote node.
An input session where a remote node-ID is specified is called a *selective session*;
such a session accepts transfers from a particular remote node-ID only.
Selective sessions are useful for service transfers.

From the above description it is easy to see that a set of transfers that are valid for a given
selective session is a subset of transfers that are valid for a given promiscuous session
sharing the same data specifier.
For example, consider two sessions sharing a data specifier *D*,
one of which is promiscuous and the other is selective bound to remote node-ID *N*.
Suppose that a transfer matching the data specifier *D* is received by the local node from remote node *N*,
thereby matching both sessions.
In cases like this,
**the transport implementation is required to deliver the received transfer into both matching sessions**.
The order (whether selective or promiscuous is served first) is implementation-defined.


Sniffing/snooping and tracing
+++++++++++++++++++++++++++++

Set up live capture on a transport using :meth:`Transport.begin_capture`.
We are using the loopback transport here for demonstration but other transports follow the same interface:

>>> from pyuavcan.transport import Capture
>>> from pyuavcan.transport.loopback import LoopbackTransport
>>> captured_events = []
>>> def on_capture(cap: Capture) -> None:
...     captured_events.append(cap)
>>> tr = LoopbackTransport(None)
>>> tr.begin_capture(on_capture)

Multiple different transports can be set up to deliver capture events into the same handler since they all
share the same transport-agnostic API.
This way, heterogeneous redundant transports can write and parse a single shared log file.

Emit a random transfer and see it captured:

>>> from pyuavcan.transport import MessageDataSpecifier, PayloadMetadata, OutputSessionSpecifier, Transfer
>>> from pyuavcan.transport import Timestamp, Priority
>>> import asyncio
>>> await_ = asyncio.get_event_loop().run_until_complete
>>> ses = tr.get_output_session(OutputSessionSpecifier(MessageDataSpecifier(1234), None), PayloadMetadata(1024))
>>> await_(ses.send(Transfer(Timestamp.now(), Priority.LOW, 1234567890, [memoryview(b'abc')]),
...                 monotonic_deadline=asyncio.get_event_loop().time() + 1.0))
True
>>> captured_events
[LoopbackCapture(...priority=LOW, transfer_id=1234567890...)]

The captured events can be processed afterwards: logged, displayed, or reconstructed into high-level events.
The latter is done with the help of :class:`Tracer` instantiated using the static factory method
:meth:`Transport.make_tracer`:

>>> tracer = LoopbackTransport.make_tracer()
>>> tracer.update(captured_events[0])  # Captures could be read from live network or from a log file, for instance.
TransferTrace(...priority=LOW, transfer_id=1234567890...)


Implementing new transports
+++++++++++++++++++++++++++

New transports can be added trivially by subclassing :class:`pyuavcan.transport.Transport`.
This module contains several nested submodules providing standard transport implementations
according to the UAVCAN specification (e.g., the UAVCAN/CAN transport) alongside with experimental implementations.

Each specific transport implementation included in the library shall reside in its own separate
submodule under :mod:`pyuavcan.transport`.
The name of the submodule should be the lowercase name of the transport.
The name of the implementation class that inherits from :class:`pyuavcan.transport.Transport`
should begin with capitalized name of the submodule followed by ``Transport``.
If the new transport contains a media sub-layer, the media interface class should be at
``pyuavcan.transport.*.media.Media``, where the asterisk is the transport name placeholder;
the media sub-layer should follow the same organization patterns as the transport layer.
See the UAVCAN/CAN transport as an example.

Implementations included in the library are never auto-imported, nor do they need to be.
The same should be true for transport-specific media sub-layers.
The application is required to explicitly import the transport (and media sub-layer) implementations that are needed.
A highly generic, transport-agnostic application may benefit from the helper functions available in
:mod:`pyuavcan.util`, designed specifically to ease discovery and use of entities defined in submodules that
are not auto-imported and whose names are not known in advance.

Users can define their custom transports and/or media sub-layers outside of the library scope.
The library itself does not care about the location of its components.


Class inheritance diagram
+++++++++++++++++++++++++

Below is the class inheritance diagram for this module (trivial classes may be omitted):

.. inheritance-diagram:: pyuavcan.transport._transport
                         pyuavcan.transport._error
                         pyuavcan.transport._session
                         pyuavcan.transport._data_specifier
                         pyuavcan.transport._transfer
                         pyuavcan.transport._payload_metadata
                         pyuavcan.transport._tracer
   :parts: 1
"""

# Please keep the imports well-ordered because it affects the generated documentation.

# Core transport.
from ._transport import Transport as Transport
from ._transport import ProtocolParameters as ProtocolParameters
from ._transport import TransportStatistics as TransportStatistics

# Transport model auxiliaries.
from ._transfer import Transfer as Transfer
from ._transfer import TransferFrom as TransferFrom
from ._transfer import Priority as Priority

from ._data_specifier import DataSpecifier as DataSpecifier
from ._data_specifier import MessageDataSpecifier as MessageDataSpecifier
from ._data_specifier import ServiceDataSpecifier as ServiceDataSpecifier

from ._session import SessionSpecifier as SessionSpecifier
from ._session import InputSessionSpecifier as InputSessionSpecifier
from ._session import OutputSessionSpecifier as OutputSessionSpecifier
from ._session import Session as Session
from ._session import InputSession as InputSession
from ._session import OutputSession as OutputSession

from ._payload_metadata import PayloadMetadata as PayloadMetadata

# Low-level entities.
from ._session import SessionStatistics as SessionStatistics
from ._session import Feedback as Feedback

from ._timestamp import Timestamp as Timestamp

from ._transfer import FragmentedPayload as FragmentedPayload

# Exceptions.
from ._error import TransportError as TransportError
from ._error import UnsupportedSessionConfigurationError as UnsupportedSessionConfigurationError
from ._error import OperationNotDefinedForAnonymousNodeError as OperationNotDefinedForAnonymousNodeError
from ._error import InvalidTransportConfigurationError as InvalidTransportConfigurationError
from ._error import InvalidMediaConfigurationError as InvalidMediaConfigurationError
from ._error import ResourceClosedError as ResourceClosedError

# Analysis API.
from ._tracer import Capture as Capture
from ._tracer import CaptureCallback as CaptureCallback
from ._tracer import AlienSessionSpecifier as AlienSessionSpecifier
from ._tracer import AlienTransferMetadata as AlienTransferMetadata
from ._tracer import AlienTransfer as AlienTransfer
from ._tracer import Trace as Trace
from ._tracer import ErrorTrace as ErrorTrace
from ._tracer import TransferTrace as TransferTrace
from ._tracer import Tracer as Tracer

# Reusable components.
from . import commons as commons



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import asyncio
import logging
import dataclasses
import pyuavcan.dsdl
import pyuavcan.transport
import pyuavcan.util
from ._base import ServiceClass, ServicePort, PortFinalizer, DEFAULT_SERVICE_REQUEST_TIMEOUT
from ._error import PortClosedError


# Shouldn't be too large as this value defines how quickly the serving task will detect that the underlying
# transport is closed.
_LISTEN_FOREVER_TIMEOUT = 1


OutputTransportSessionFactory = typing.Callable[[int], pyuavcan.transport.OutputSession]
ServiceRequestClass = typing.TypeVar("ServiceRequestClass", bound=pyuavcan.dsdl.CompositeObject)
ServiceResponseClass = typing.TypeVar("ServiceResponseClass", bound=pyuavcan.dsdl.CompositeObject)


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ServerStatistics:
    request_transport_session: pyuavcan.transport.SessionStatistics
    """There is only one input transport session per server."""

    response_transport_sessions: typing.Dict[int, pyuavcan.transport.SessionStatistics]
    """This is a mapping keyed by the remote client node-ID value. One transport session per client."""

    served_requests: int

    deserialization_failures: int
    """Requests that could not be received because of bad input transfers."""

    malformed_requests: int
    """Problems at the transport layer."""


@dataclasses.dataclass(frozen=True)
class ServiceRequestMetadata:
    """
    This structure is supplied with every received request for informational purposes.
    The application is not required to do anything with it.
    """

    timestamp: pyuavcan.transport.Timestamp
    """Timestamp of the first frame of the request transfer."""

    priority: pyuavcan.transport.Priority
    """Same priority will be used for the response (see Specification)."""

    transfer_id: int
    """Same transfer-ID will be used for the response (see Specification)."""

    client_node_id: int
    """The response will be sent back to this node."""

    def __repr__(self) -> str:
        kwargs = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        kwargs["priority"] = self.priority.name
        del kwargs["timestamp"]
        return pyuavcan.util.repr_attributes(self, str(self.timestamp), **kwargs)


ServiceRequestHandler = typing.Callable[
    [ServiceRequestClass, ServiceRequestMetadata], typing.Awaitable[typing.Optional[ServiceResponseClass]]
]
"""Type of the async request handler callable."""


class Server(ServicePort[ServiceClass]):
    """
    At most one task can use the server at any given time.
    The instance must be closed manually to stop the server.

    .. note::
        Normally we should use correct generic types ``ServiceClass.Request`` and ``ServiceClass.Response`` in the API;
        however, MyPy does not support that yet. Please find the context at
        https://github.com/python/mypy/issues/7121 (please upvote!) and https://github.com/UAVCAN/pyuavcan/issues/61.
        We use a tentative workaround for now to silence bogus type errors. When the missing logic is implemented
        in MyPy, this should be switched back to proper implementation.
    """

    def __init__(
        self,
        dtype: typing.Type[ServiceClass],
        input_transport_session: pyuavcan.transport.InputSession,
        output_transport_session_factory: OutputTransportSessionFactory,
        finalizer: PortFinalizer,
        loop: asyncio.AbstractEventLoop,
    ):
        """
        Do not call this directly! Use :meth:`Presentation.get_server`.
        """
        self._dtype = dtype
        self._input_transport_session = input_transport_session
        self._output_transport_session_factory = output_transport_session_factory
        self._finalizer = finalizer
        self._loop = loop

        self._output_transport_sessions: typing.Dict[int, pyuavcan.transport.OutputSession] = {}
        self._maybe_task: typing.Optional[asyncio.Task[None]] = None
        self._closed = False
        self._send_timeout = DEFAULT_SERVICE_REQUEST_TIMEOUT

        self._served_request_count = 0
        self._deserialization_failure_count = 0
        self._malformed_request_count = 0

    # ----------------------------------------  MAIN API  ----------------------------------------

    async def serve(
        self,
        handler: ServiceRequestHandler[ServiceRequestClass, ServiceResponseClass],
        monotonic_deadline: typing.Optional[float] = None,
    ) -> None:
        """
        This is like :meth:`serve_for` except that it exits normally after the specified monotonic deadline is reached.
        The deadline value is compared against :meth:`asyncio.AbstractEventLoop.time`.
        If no deadline is provided, it is assumed to be infinite.
        """
        # Observe that if we aggregate redundant transports with different non-monotonic transfer ID modulo values,
        # it might be that the transfer ID that we obtained from the request may be invalid for some of the transports.
        # This is why we can't reliably aggregate redundant transports with different transfer-ID overflow parameters.
        while not self._closed:
            out: typing.Optional[typing.Tuple[pyuavcan.dsdl.CompositeObject, ServiceRequestMetadata]]
            if monotonic_deadline is None:
                out = await self._receive(self._loop.time() + _LISTEN_FOREVER_TIMEOUT)
                if out is None:
                    continue
            else:
                out = await self._receive(monotonic_deadline)
                if out is None:
                    break  # Timed out.

            self._served_request_count += 1
            request, meta = out
            response: typing.Optional[ServiceResponseClass] = None  # Fallback state
            assert isinstance(request, self._dtype.Request), "Internal protocol violation"
            try:
                response = await handler(request, meta)  # type: ignore
                if response is not None and not isinstance(response, self._dtype.Response):
                    raise TypeError(
                        f"The application request handler has returned an invalid response: "
                        f"expected an instance of {self._dtype.Response} or None, "
                        f"found {type(response)} instead. "
                        f"The corresponding request was {request} with metadata {meta}."
                    )
            except Exception as ex:
                if isinstance(ex, asyncio.CancelledError):
                    raise
                _logger.exception("%s unhandled exception in the handler: %s", self, ex)

            response_transport_session = self._get_output_transport_session(meta.client_node_id)

            # Send the response unless the application has opted out, in which case do nothing.
            if response is not None:
                # TODO: make the send timeout configurable.
                await self._do_send(response, meta, response_transport_session, self._loop.time() + self._send_timeout)

    async def serve_for(
        self, handler: ServiceRequestHandler[ServiceRequestClass, ServiceResponseClass], timeout: float
    ) -> None:
        """
        Listen for requests for the specified time or until the instance is closed, then exit.

        When a request is received, the supplied handler callable will be invoked with the request object
        and the associated metadata object (which contains auxiliary information such as the client's node-ID).
        The handler shall return the response or None. If None is returned, the server will not send any response back
        (this practice is discouraged). If the handler throws an exception, it will be suppressed and logged.
        """
        return await self.serve(handler, monotonic_deadline=self._loop.time() + timeout)

    def serve_in_background(self, handler: ServiceRequestHandler[ServiceRequestClass, ServiceResponseClass]) -> None:
        """
        Start a new task and use it to run the server in the background.
        The task will be stopped when the server is closed.

        When a request is received, the supplied handler callable will be invoked with the request object
        and the associated metadata object (which contains auxiliary information such as the client's node-ID).
        The handler shall return the response or None. If None is returned, the server will not send any response back
        (this practice is discouraged). If the handler throws an exception, it will be suppressed and logged.

        If the background task is already running, it will be cancelled and a new one will be started instead.
        This method of serving requests shall not be used concurrently with other methods.
        """

        async def task_function() -> None:
            while not self._closed:
                try:
                    await self.serve_for(handler, _LISTEN_FOREVER_TIMEOUT)
                except asyncio.CancelledError:
                    _logger.debug("%s task cancelled", self)
                    break
                except pyuavcan.transport.ResourceClosedError as ex:
                    _logger.debug("%s task got a resource closed error and will exit: %s", self, ex)
                    break
                except Exception as ex:
                    _logger.exception("%s task failure: %s", self, ex)
                    await asyncio.sleep(1)  # TODO is this an adequate failure management strategy?

        if self._maybe_task is not None:
            self._maybe_task.cancel()

        self._raise_if_closed()
        self._maybe_task = self._loop.create_task(task_function())

    # ----------------------------------------  AUXILIARY  ----------------------------------------

    @property
    def send_timeout(self) -> float:
        """
        Every response transfer will have to be sent in this amount of time.
        If the time is exceeded, the attempt is aborted and a warning is logged.
        The default value is :data:`DEFAULT_SERVICE_REQUEST_TIMEOUT`.
        """
        return self._send_timeout

    @send_timeout.setter
    def send_timeout(self, value: float) -> None:
        value = float(value)
        if 0 < value < float("+inf"):
            self._send_timeout = value
        else:
            raise ValueError(f"Invalid send timeout value: {value}")

    def sample_statistics(self) -> ServerStatistics:
        """
        Returns the statistical counters of this server instance,
        including the statistical metrics of the underlying transport sessions.
        """
        return ServerStatistics(
            request_transport_session=self._input_transport_session.sample_statistics(),
            response_transport_sessions={
                nid: ts.sample_statistics() for nid, ts in self._output_transport_sessions.items()
            },
            served_requests=self._served_request_count,
            deserialization_failures=self._deserialization_failure_count,
            malformed_requests=self._malformed_request_count,
        )

    @property
    def dtype(self) -> typing.Type[ServiceClass]:
        return self._dtype

    @property
    def input_transport_session(self) -> pyuavcan.transport.InputSession:
        return self._input_transport_session

    def close(self) -> None:
        if not self._closed:
            self._closed = True
            if self._maybe_task is not None:  # The task may be holding the lock.
                try:
                    self._maybe_task.cancel()  # We don't wait for it to exit because it's pointless.
                except Exception as ex:
                    _logger.exception("%s task could not be cancelled: %s", self, ex)
                self._maybe_task = None

            self._finalizer((self._input_transport_session, *self._output_transport_sessions.values()))

    async def _receive(
        self, monotonic_deadline: float
    ) -> typing.Optional[typing.Tuple[ServiceRequestClass, ServiceRequestMetadata]]:
        while True:
            transfer = await self._input_transport_session.receive(monotonic_deadline)
            if transfer is None:
                return None
            if transfer.source_node_id is not None:
                meta = ServiceRequestMetadata(
                    timestamp=transfer.timestamp,
                    priority=transfer.priority,
                    transfer_id=transfer.transfer_id,
                    client_node_id=transfer.source_node_id,
                )
                request = pyuavcan.dsdl.deserialize(self._dtype.Request, transfer.fragmented_payload)
                if request is not None:
                    return request, meta  # type: ignore
                self._deserialization_failure_count += 1
            else:
                self._malformed_request_count += 1

    @staticmethod
    async def _do_send(
        response: ServiceResponseClass,
        metadata: ServiceRequestMetadata,
        session: pyuavcan.transport.OutputSession,
        monotonic_deadline: float,
    ) -> bool:
        timestamp = pyuavcan.transport.Timestamp.now()
        fragmented_payload = list(pyuavcan.dsdl.serialize(response))
        transfer = pyuavcan.transport.Transfer(
            timestamp=timestamp,
            priority=metadata.priority,
            transfer_id=metadata.transfer_id,
            fragmented_payload=fragmented_payload,
        )
        return await session.send(transfer, monotonic_deadline)

    def _get_output_transport_session(self, client_node_id: int) -> pyuavcan.transport.OutputSession:
        try:
            return self._output_transport_sessions[client_node_id]
        except LookupError:
            out = self._output_transport_session_factory(client_node_id)
            self._output_transport_sessions[client_node_id] = out
            return out

    def _raise_if_closed(self) -> None:
        if self._closed:
            raise PortClosedError(repr(self))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import copy
import typing
import asyncio
import logging
import dataclasses
import pyuavcan
from pyuavcan.transport import Timestamp
from pyuavcan.transport.commons.high_overhead_transport import TransferReassembler
from .._frame import UDPFrame


_logger = logging.getLogger(__name__)


class UDPInputSessionStatistics(pyuavcan.transport.SessionStatistics):
    pass


class UDPInputSession(pyuavcan.transport.InputSession):
    """
    As you already know, the UDP port number is a function of the data specifier.
    Hence, the input flow demultiplexing is mostly done by the UDP/IP stack implemented in the operating system
    itself, we just need to put a few basic abstractions on top.
    One of those abstractions is the internal socket reader class, which is not part of the API
    but its function is important if one needs to understand how the data flow is organized inside the library::

        [Socket] 1   --->   1 [Demultiplexer] 1   --->   * [Input session] 1   --->   1 [API]

    *(The plurality notation is supposed to resemble UML: 1 - one, * - many.)*

    A UDP datagram is an atomic unit of workload for the stack.
    Unlike, say, the serial transport, the operating system does the low-level work of framing and
    CRC checking for us (thank you), so we get our stuff sorted up to the OSI layer 4 inclusive.
    The processing pipeline per datagram is as follows:

    - The socket reader obtains the datagram from the socket using ``recvfrom()``.
      The source IP address is mapped to a node-ID and the contents are parsed into a UAVCAN UDP frame instance.
      If anything goes wrong here (like if the source IP address belongs to a wrong subnet or the datagram
      does not contain a valid UAVCAN frame or whatever), the datagram is dropped and the appropriate statistical
      counters are updated.

    - The socket reader looks up the input session instances that have subscribed for the datagram from the
      current source node-ID (derived from the IP address) and passes the frame to them.
      By the way, remember that this is a zero-copy stack, so every subscribed input session gets a reference
      to the same instance of the frame, although it is beside the point right now.

    - Upon reception of the frame, the input session (one of many) updates its reassembler state machine
      and runs all that meticulous bookkeeping you can't get away from if you need to receive multi-frame transfers.

    - If the received frame happened to complete a transfer, the input session passes it up to the higher layer.

    Now, an attentive reader might exclaim:

        But look! If there is more than one input session instance per source node-ID,
        we'd be running multiple transfer reassemblers with the same input data,
        which is inefficient!
        Why can't we extract the task of transfer reassembly into the socket reader,
        before the pipeline is forked, to avoid the extra workload?

    That is a good question, and here's why:

    - The most important reason is that the proposal would only work if the state of
      a transfer reassembler was a function of the input frame flow only.
      This is not the case.
      The state of a transfer reassembler is also defined by its configuration parameters
      which are defined per-instance, which in turn are defined per input session instance.
      In particular, the transfer-ID timeout is configured separately per input session.

    - The case where there is more than one input session per remote node-ID is uncommon.
      In fact, it may only occur if the higher layers requested a promiscuous and a selective session
      at the same time, which normally does not happen with UAVCAN.
      We support this use case nevertheless because this library is supposed to offer a generic and
      flexible API due to its intended usages (read the library design goals).

    - The computing load of updating the state machine of a transfer reassembler is minuscule.
      The most intensive computation happening there is the CRC update, which is not intense at all.

    The architecture of the data processing pipeline in PyUAVCAN is complex, but that is due to the
    high-level requirements for the library: it has to support *all transport protocols*, a lot of
    media layers, and be portable, so trade-offs had to be made.
    It should be understood that actual safety-critical implementations used in production systems
    can be far simpler because generally they do not have to be multi-transport and multi-platform.
    """

    DEFAULT_TRANSFER_ID_TIMEOUT = 2.0
    """
    Units are seconds. Can be overridden after instantiation if needed.
    """

    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        finalizer: typing.Callable[[], None],
    ):
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._loop = loop
        self._maybe_finalizer: typing.Optional[typing.Callable[[], None]] = finalizer
        assert isinstance(self._specifier, pyuavcan.transport.InputSessionSpecifier)
        assert isinstance(self._payload_metadata, pyuavcan.transport.PayloadMetadata)
        assert isinstance(self._loop, asyncio.AbstractEventLoop)
        assert callable(self._maybe_finalizer)

        self._transfer_id_timeout = self.DEFAULT_TRANSFER_ID_TIMEOUT
        self._queue: asyncio.Queue[pyuavcan.transport.TransferFrom] = asyncio.Queue()

    def _process_frame(self, timestamp: Timestamp, source_node_id: int, frame: typing.Optional[UDPFrame]) -> None:
        """
        The source node-ID is always valid because anonymous transfers are not defined for the UDP transport.
        The frame argument may be None to indicate that the underlying transport has received a datagram
        which is valid but does not contain a UAVCAN UDP frame inside. This is needed for error stats tracking.

        This is a part of the transport-internal API. It's a public method despite the name because Python's
        visibility handling capabilities are limited. I guess we could define a private abstract base to
        handle this but it feels like too much work. Why can't we have protected visibility in Python?
        """
        assert isinstance(source_node_id, int) and source_node_id >= 0, "Internal protocol violation"
        if frame is None:  # Malformed frame.
            self._statistics.errors += 1
            return
        self._statistics.frames += 1

        transfer = self._get_reassembler(source_node_id).process_frame(timestamp, frame, self._transfer_id_timeout)
        if transfer is not None:
            self._statistics.transfers += 1
            self._statistics.payload_bytes += sum(map(len, transfer.fragmented_payload))
            _logger.debug("%s: Received transfer: %s; current stats: %s", self, transfer, self._statistics)
            try:
                self._queue.put_nowait(transfer)
            except asyncio.QueueFull:  # pragma: no cover
                # TODO: make the queue capacity configurable
                self._statistics.drops += len(transfer.fragmented_payload)

    async def receive(self, monotonic_deadline: float) -> typing.Optional[pyuavcan.transport.TransferFrom]:
        try:
            timeout = monotonic_deadline - self._loop.time()
            if timeout > 0:
                transfer = await asyncio.wait_for(self._queue.get(), timeout)
            else:
                transfer = self._queue.get_nowait()
        except (asyncio.TimeoutError, asyncio.QueueEmpty):
            # If there are unprocessed transfers, allow the caller to read them even if the instance is closed.
            if self._maybe_finalizer is None:
                raise pyuavcan.transport.ResourceClosedError(f"{self} is closed") from None
            return None
        else:
            assert isinstance(transfer, pyuavcan.transport.TransferFrom), "Internal protocol violation"
            assert transfer.source_node_id == self._specifier.remote_node_id or self._specifier.remote_node_id is None
            return transfer

    @property
    def transfer_id_timeout(self) -> float:
        return self._transfer_id_timeout

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        if value > 0:
            self._transfer_id_timeout = float(value)
        else:
            raise ValueError(f"Invalid value for transfer-ID timeout [second]: {value}")

    @property
    def specifier(self) -> pyuavcan.transport.InputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def close(self) -> None:
        if self._maybe_finalizer is not None:
            self._maybe_finalizer()
            self._maybe_finalizer = None

    @property
    @abc.abstractmethod
    def _statistics(self) -> UDPInputSessionStatistics:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_reassembler(self, source_node_id: int) -> TransferReassembler:
        raise NotImplementedError


@dataclasses.dataclass
class PromiscuousUDPInputSessionStatistics(UDPInputSessionStatistics):
    reassembly_errors_per_source_node_id: typing.Dict[
        int, typing.Dict[TransferReassembler.Error, int]
    ] = dataclasses.field(default_factory=dict)
    """
    Keys are source node-IDs; values are dicts where keys are error enum members and values are counts.
    """


class PromiscuousUDPInputSession(UDPInputSession):
    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly, use the factory method instead.
        """
        self._statistics_impl = PromiscuousUDPInputSessionStatistics()
        self._reassemblers: typing.Dict[int, TransferReassembler] = {}
        super().__init__(specifier=specifier, payload_metadata=payload_metadata, loop=loop, finalizer=finalizer)

    def sample_statistics(self) -> PromiscuousUDPInputSessionStatistics:
        return copy.copy(self._statistics)

    @property
    def _statistics(self) -> PromiscuousUDPInputSessionStatistics:
        return self._statistics_impl

    def _get_reassembler(self, source_node_id: int) -> TransferReassembler:
        assert isinstance(source_node_id, int) and source_node_id >= 0, "Internal protocol violation"
        try:
            return self._reassemblers[source_node_id]
        except LookupError:

            def on_reassembly_error(error: TransferReassembler.Error) -> None:
                self._statistics.errors += 1
                d = self._statistics.reassembly_errors_per_source_node_id[source_node_id]
                try:
                    d[error] += 1
                except LookupError:
                    d[error] = 1

            self._statistics.reassembly_errors_per_source_node_id.setdefault(source_node_id, {})
            reasm = TransferReassembler(
                source_node_id=source_node_id,
                extent_bytes=self._payload_metadata.extent_bytes,
                on_error_callback=on_reassembly_error,
            )
            self._reassemblers[source_node_id] = reasm
            _logger.debug("%s: New %s (%d total)", self, reasm, len(self._reassemblers))
            return reasm


@dataclasses.dataclass
class SelectiveUDPInputSessionStatistics(UDPInputSessionStatistics):
    reassembly_errors: typing.Dict[TransferReassembler.Error, int] = dataclasses.field(default_factory=dict)
    """
    Keys are error enum members and values are counts.
    """


class SelectiveUDPInputSession(UDPInputSession):
    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        loop: asyncio.AbstractEventLoop,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly, use the factory method instead.
        """
        self._statistics_impl = SelectiveUDPInputSessionStatistics()

        source_node_id = specifier.remote_node_id
        assert source_node_id is not None, "Internal protocol violation"

        def on_reassembly_error(error: TransferReassembler.Error) -> None:
            self._statistics.errors += 1
            try:
                self._statistics.reassembly_errors[error] += 1
            except LookupError:
                self._statistics.reassembly_errors[error] = 1

        self._reassembler = TransferReassembler(
            source_node_id=source_node_id,
            extent_bytes=payload_metadata.extent_bytes,
            on_error_callback=on_reassembly_error,
        )

        super().__init__(specifier=specifier, payload_metadata=payload_metadata, loop=loop, finalizer=finalizer)

    def sample_statistics(self) -> SelectiveUDPInputSessionStatistics:
        return copy.copy(self._statistics)

    @property
    def _statistics(self) -> SelectiveUDPInputSessionStatistics:
        return self._statistics_impl

    def _get_reassembler(self, source_node_id: int) -> TransferReassembler:
        assert source_node_id == self._reassembler.source_node_id, "Internal protocol violation"
        return self._reassembler



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import asyncio
import logging
import dataclasses
from typing import Optional, Set, Any
import pydsdl
import pyuavcan.application
from pyuavcan.transport import MessageDataSpecifier, ServiceDataSpecifier
from uavcan.node.port import List_0_1 as List
from uavcan.node.port import SubjectIDList_0_1 as SubjectIDList
from uavcan.node.port import ServiceIDList_0_1 as ServiceIDList
from uavcan.node.port import SubjectID_1_0 as SubjectID


@dataclasses.dataclass(frozen=True)
class _State:
    pub: Set[int]
    sub: Set[int]
    cln: Set[int]
    srv: Set[int]


class PortListPublisher:
    """
    This class is to be automatically instantiated by :class:`pyuavcan.application.Node`.
    Publishing will be suspended while the local node-ID is anonymous.
    The status is updated every second, publications happen every MAX_PUBLICATION_PERIOD seconds or on change.
    """

    _UPDATE_PERIOD = 1.0
    _MAX_UPDATES_BETWEEN_PUBLICATIONS = int(List.MAX_PUBLICATION_PERIOD / _UPDATE_PERIOD)

    def __init__(self, node: pyuavcan.application.Node) -> None:
        self._node = node
        self._pub: Optional[pyuavcan.presentation.Publisher[List]] = None
        self._updates_since_pub = 0
        self._next_update_at = 0.0
        self._timer: Optional[asyncio.TimerHandle] = None
        self._state = _State(set(), set(), set(), set())

        def start() -> None:
            self._next_update_at = self.node.loop.time() + PortListPublisher._UPDATE_PERIOD
            self._timer = self.node.loop.call_at(self._next_update_at, self._update)

        def close() -> None:
            if self._pub is not None:
                self._pub.close()
            if self._timer is not None:
                self._timer.cancel()
                self._timer = None

        self.node.add_lifetime_hooks(start, close)

    @property
    def node(self) -> pyuavcan.application.Node:
        return self._node

    def _get_publisher(self) -> Optional[pyuavcan.presentation.Publisher[List]]:
        if self._pub is None:
            try:
                self._pub = self.node.make_publisher(List)
                self._pub.priority = pyuavcan.transport.Priority.OPTIONAL
            except Exception as ex:  # pragma: no cover
                _logger.exception("%r: Could not initialize the publisher: %s", self, ex)
            else:
                _logger.debug("%r: Publisher initialized: %r", self, self._pub)
        return self._pub

    def _update(self) -> None:
        self._updates_since_pub += 1
        self._next_update_at += PortListPublisher._UPDATE_PERIOD
        self._timer = self.node.loop.call_at(self._next_update_at, self._update)

        if self.node.id is None:
            return
        publisher = self._get_publisher()
        if publisher is None:
            return

        trans = self.node.presentation.transport
        input_ds = [x.specifier.data_specifier for x in trans.input_sessions]
        srv_in_ds = [x for x in input_ds if isinstance(x, ServiceDataSpecifier)]
        state = _State(
            pub={
                x.specifier.data_specifier.subject_id
                for x in trans.output_sessions
                if isinstance(x.specifier.data_specifier, MessageDataSpecifier)
            },
            sub={x.subject_id for x in input_ds if isinstance(x, MessageDataSpecifier)},
            cln={x.service_id for x in srv_in_ds if x.role == ServiceDataSpecifier.Role.RESPONSE},
            srv={x.service_id for x in srv_in_ds if x.role == ServiceDataSpecifier.Role.REQUEST},
        )

        state_changed = state != self._state
        time_expired = self._updates_since_pub >= PortListPublisher._MAX_UPDATES_BETWEEN_PUBLICATIONS
        if state_changed or time_expired:
            _logger.debug("%r: Publishing: state_changed=%r, state=%r", self, state_changed, state)
            self._state = state
            self._updates_since_pub = 0  # Should we handle ResourceClosedError here?
            publisher.publish_soon(_make_port_list(self._state, trans.capture_active))

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self.node)


_logger = logging.getLogger(__name__)


def _make_port_list(state: _State, packet_capture_mode: bool) -> List:
    from uavcan.primitive import Empty_1_0 as Empty

    return List(
        publishers=_make_subject_id_list(state.pub),
        subscribers=_make_subject_id_list(state.sub) if not packet_capture_mode else SubjectIDList(total=Empty()),
        clients=_make_service_id_list(state.cln),
        servers=_make_service_id_list(state.srv),
    )


def _make_subject_id_list(ports: Set[int]) -> SubjectIDList:
    sparse_list_type = pyuavcan.dsdl.get_model(SubjectIDList)["sparse_list"].data_type
    assert isinstance(sparse_list_type, pydsdl.ArrayType)

    if len(ports) <= sparse_list_type.capacity:
        return SubjectIDList(sparse_list=[SubjectID(x) for x in sorted(ports)])

    out = SubjectIDList()
    assert out.mask is not None
    _populate_mask(ports, out.mask)
    return out


def _make_service_id_list(ports: Set[int]) -> ServiceIDList:
    out = ServiceIDList()
    _populate_mask(ports, out.mask)
    return out


def _populate_mask(ports: Set[int], output: Any) -> None:
    for idx in range(len(output)):  # pylint: disable=consider-using-enumerate
        output[idx] = idx in ports


def _unittest_make_port_list() -> None:
    state = _State(
        pub={1, 8191, 0},
        sub=set(range(257)),
        cln=set(),
        srv=set(range(512)),
    )

    msg = _make_port_list(state, False)

    assert msg.publishers.sparse_list is not None
    pubs = [x.value for x in msg.publishers.sparse_list]
    assert pubs == [0, 1, 8191]  # Sorted!

    assert msg.subscribers.mask is not None
    assert msg.subscribers.mask.sum() == 257
    for idx in range(SubjectIDList.CAPACITY):
        assert msg.subscribers.mask[idx] == (idx < 257)

    assert msg.clients.mask.sum() == 0
    assert msg.servers.mask.sum() == 512


def _unittest_populate_mask() -> None:
    srv = SubjectIDList()
    mask = srv.mask
    assert mask is not None
    _populate_mask({1, 2, 8191}, mask)
    for idx in range(SubjectIDList.CAPACITY):
        assert mask[idx] == (idx in {1, 2, 8191})



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import pyuavcan


class SerialSession:
    def __init__(self, finalizer: typing.Callable[[], None]):
        self._close_finalizer: typing.Optional[typing.Callable[[], None]] = finalizer
        if not callable(self._close_finalizer):  # pragma: no cover
            raise TypeError(f"Invalid finalizer: {type(self._close_finalizer).__name__}")

    def close(self) -> None:
        fin = self._close_finalizer
        if fin is not None:
            self._close_finalizer = None
            fin()

    def _raise_if_closed(self) -> None:
        if self._close_finalizer is None:
            raise pyuavcan.transport.ResourceClosedError(f"Session is closed: {self}")



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import asyncio
import pyuavcan.util
from pyuavcan.transport import Timestamp
from ._frame import Envelope
from ._filter import FilterConfiguration


class Media(abc.ABC):
    """
    CAN hardware abstraction interface.

    It is recognized that the availability of some of the media implementations may be conditional on the type of
    platform (e.g., SocketCAN is Linux-only) and the availability of third-party software (e.g., PySerial may be
    needed for SLCAN). Python packages containing such media implementations shall be always importable.
    """

    ReceivedFramesHandler = typing.Callable[[typing.Sequence[typing.Tuple[Timestamp, Envelope]]], None]
    """
    The frames handler is non-blocking and non-yielding; returns immediately.
    The timestamp is provided individually per frame.
    """

    VALID_MTU_SET = {8, 12, 16, 20, 24, 32, 48, 64}
    """Valid MTU values for Classic CAN and CAN FD."""

    @property
    @abc.abstractmethod
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        The asyncio event loop used to operate the media instance.
        Shall be the same one as that of the parent transport.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def interface_name(self) -> str:
        """
        The name of the interface on the local system. For example:

        - ``can0`` for SocketCAN;
        - ``/dev/serial/by-id/usb-Zubax_Robotics_Zubax_Babel_28002E0001514D593833302000000000-if00`` for SLCAN;
        - ``COM9`` for SLCAN.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def mtu(self) -> int:
        """
        The value belongs to :attr:`VALID_MTU_SET`.
        Observe that the media interface doesn't care whether we're using CAN FD or CAN 2.0 because the UAVCAN
        CAN transport protocol itself doesn't care. The transport simply does not distinguish them.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def number_of_acceptance_filters(self) -> int:
        """
        The number of hardware acceptance filters supported by the underlying CAN controller.
        Some media drivers, such as SocketCAN, may implement acceptance filtering in software instead of hardware.
        The returned value shall be a positive integer. If the hardware does not support filtering at all,
        the media driver shall emulate at least one filter in software.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def start(self, handler: ReceivedFramesHandler, no_automatic_retransmission: bool) -> None:
        """
        Every received frame shall be timestamped. Both monotonic and system timestamps are required.
        There are no timestamping accuracy requirements. An empty set of frames should never be reported.

        The media implementation shall drop all non-data frames (RTR frames, error frames, etc.).

        If the set contains more than one frame, all frames must be ordered by the time of their arrival,
        which also should be reflected in their timestamps; that is, the timestamp of a frame at index N
        generally should not be higher than the timestamp of a frame at index N+1. The timestamp ordering,
        however, is not a strict requirement because it is recognized that due to error variations in the
        timestamping algorithms timestamp values may not be monotonically increasing.

        The implementation should strive to return as many frames per call as possible as long as that
        does not increase the worst case latency.

        The handler shall be invoked on the event loop returned by :attr:`loop`.

        The transport is guaranteed to invoke this method exactly once during (or shortly after) initialization;
        it can be used to perform a lazy start of the receive loop task/thread/whatever.
        It is undefined behavior to invoke this method more than once on the same instance.

        :param handler: Behold my transformation. You are empowered to do as you please.

        :param no_automatic_retransmission: If True, the CAN controller should be configured to abort transmission
            of CAN frames after first error or arbitration loss (time-triggered transmission mode).
            This mode is used by UAVCAN to facilitate the PnP node-ID allocation process on the client side.
            Its support is not mandatory but highly recommended to avoid excessive disturbance of the bus
            while PnP allocations are in progress.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def configure_acceptance_filters(self, configuration: typing.Sequence[FilterConfiguration]) -> None:
        """
        This method is invoked whenever the subscription set is changed in order to communicate to the underlying
        CAN controller hardware which CAN frames should be accepted and which ones should be ignored.

        An empty set of configurations means that the transport is not interested in any frames, i.e., all frames
        should be rejected by the controller. That is also the recommended default configuration (ignore all frames
        until explicitly requested otherwise).
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def send(self, frames: typing.Iterable[Envelope], monotonic_deadline: float) -> int:
        """
        All passed frames are guaranteed to share the same CAN-ID. This guarantee may enable some optimizations.
        The frames shall be delivered to the bus in the same order. The iterable is guaranteed to be non-empty.

        The method returns when the deadline is reached even if some of the frames could not be transmitted.
        The returned value is the number of frames that have been sent. If the returned number is lower than
        the number of supplied frames, the outer transport logic will register an error, which is then propagated
        upwards all the way to the application level.

        The method should avoid yielding the execution flow; instead, it is recommended to unload the frames
        into an internal transmission queue and return ASAP, as that minimizes the likelihood of inner
        priority inversion. If that approach is used, implementations are advised to keep track of transmission
        deadline on a per-frame basis to meet the timing requirements imposed by the application.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """
        After the media instance is closed, none of its methods can be used anymore.
        If a method is invoked after close, :class:`pyuavcan.transport.ResourceClosedError` should be raised.
        This method is an exception to that rule: if invoked on a closed instance, it shall do nothing.
        """
        raise NotImplementedError

    @staticmethod
    def list_available_interface_names() -> typing.Iterable[str]:
        """
        Returns the list of interface names that can be used with the media class implementing it.
        For example, for the SocketCAN media class it would return the SocketCAN interface names such as "vcan0";
        for SLCAN it would return the list of serial ports.

        Implementations should strive to sort the output so that the interfaces that are most likely to be used
        are listed first -- this helps GUI applications.

        If the media implementation cannot be used on the local platform,
        the method shall return an empty set instead of raising an error.
        This guarantee supports an important use case where the caller would just iterate over all inheritors
        of this Media interface and ask each one to yield the list of available interfaces,
        and then just present that to the user.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, repr(self.interface_name), mtu=self.mtu)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._loopback import LoopbackTransport as LoopbackTransport

from ._input_session import LoopbackInputSession as LoopbackInputSession

from ._output_session import LoopbackOutputSession as LoopbackOutputSession
from ._output_session import LoopbackFeedback as LoopbackFeedback

from ._tracer import LoopbackCapture as LoopbackCapture
from ._tracer import LoopbackTracer as LoopbackTracer



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import itertools
import pyuavcan
from .._frame import UAVCANFrame, TRANSFER_CRC_LENGTH_BYTES


_PADDING_PATTERN = b"\x00"


def serialize_transfer(
    compiled_identifier: int,
    transfer_id: int,
    fragmented_payload: typing.Sequence[memoryview],
    max_frame_payload_bytes: int,
) -> typing.Iterable[UAVCANFrame]:
    """
    We never request loopback for the whole transfer since it doesn't make sense. Instead, loopback request is
    always limited to the first frame only since it's sufficient for timestamping purposes.
    """
    if max_frame_payload_bytes < 1:  # pragma: no cover
        raise ValueError(f"Invalid max payload: {max_frame_payload_bytes}")

    payload_length = sum(map(len, fragmented_payload))

    if payload_length <= max_frame_payload_bytes:  # SINGLE-FRAME TRANSFER
        if payload_length > 0:
            padding_length = UAVCANFrame.get_required_padding(payload_length)
            refragmented = pyuavcan.transport.commons.refragment(
                itertools.chain(fragmented_payload, (memoryview(_PADDING_PATTERN * padding_length),)),
                max_frame_payload_bytes,
            )
            (payload,) = tuple(refragmented)
        else:
            # The special case is necessary because refragment() yields nothing if the payload is empty
            payload = memoryview(b"")

        assert max_frame_payload_bytes >= len(payload) >= payload_length
        yield UAVCANFrame(
            identifier=compiled_identifier,
            padded_payload=payload,
            transfer_id=transfer_id,
            start_of_transfer=True,
            end_of_transfer=True,
            toggle_bit=True,
        )
    else:  # MULTI-FRAME TRANSFER
        # Compute padding
        last_frame_payload_length = payload_length % max_frame_payload_bytes
        if last_frame_payload_length + TRANSFER_CRC_LENGTH_BYTES >= max_frame_payload_bytes:
            padding = b""
        else:
            last_frame_data_length = last_frame_payload_length + TRANSFER_CRC_LENGTH_BYTES
            padding = _PADDING_PATTERN * UAVCANFrame.get_required_padding(last_frame_data_length)

        # Fragment generator that goes over the padding and CRC also
        crc_bytes = pyuavcan.transport.commons.crc.CRC16CCITT.new(*fragmented_payload, padding).value_as_bytes
        refragmented = pyuavcan.transport.commons.refragment(
            itertools.chain(fragmented_payload, (memoryview(padding + crc_bytes),)), max_frame_payload_bytes
        )

        # Serialized frame emission
        for index, (last, frag) in enumerate(pyuavcan.util.mark_last(refragmented)):
            first = index == 0
            yield UAVCANFrame(
                identifier=compiled_identifier,
                padded_payload=frag,
                transfer_id=transfer_id,
                start_of_transfer=first,
                end_of_transfer=last,
                toggle_bit=index % 2 == 0,
            )


def _unittest_can_serialize_transfer() -> None:
    from ..media import DataFrame, FrameFormat

    mv = memoryview
    meta = typing.TypeVar("meta")

    def mkf(
        identifier: int,
        data: typing.Union[bytearray, bytes],
        transfer_id: int,
        start_of_transfer: bool,
        end_of_transfer: bool,
        toggle_bit: bool,
    ) -> DataFrame:
        tail = transfer_id
        if start_of_transfer:
            tail |= 1 << 7
        if end_of_transfer:
            tail |= 1 << 6
        if toggle_bit:
            tail |= 1 << 5

        data = bytearray(data)
        data.append(tail)

        return DataFrame(identifier=identifier, data=data, format=FrameFormat.EXTENDED)

    def run(
        compiled_identifier: int,
        transfer_id: int,
        fragmented_payload: typing.Sequence[memoryview],
        max_frame_payload_bytes: int,
    ) -> typing.Iterable[DataFrame]:
        for f in serialize_transfer(
            compiled_identifier=compiled_identifier,
            transfer_id=transfer_id,
            fragmented_payload=fragmented_payload,
            max_frame_payload_bytes=max_frame_payload_bytes,
        ):
            yield f.compile()

    def one(items: typing.Iterable[meta]) -> meta:
        (out,) = list(items)
        return out

    assert mkf(0xBADC0FE, b"Hello", 0, True, True, True) == one(run(0xBADC0FE, 32, [mv(b"Hell"), mv(b"o")], 7))

    assert mkf(0xBADC0FE, bytes(range(60)) + b"\x00\x00\x00", 19, True, True, True) == one(
        run(0xBADC0FE, 32 + 19, [mv(bytes(range(60)))], 63)
    )

    crc = pyuavcan.transport.commons.crc.CRC16CCITT()
    crc.add(bytes(range(0x1E)))
    assert crc.value == 0x3554
    assert [
        mkf(0xBADC0FE, b"\x00\x01\x02\x03\x04\x05\x06", 19, True, False, True),
        mkf(0xBADC0FE, b"\x07\x08\x09\x0a\x0b\x0c\x0d", 19, False, False, False),
        mkf(0xBADC0FE, b"\x0e\x0f\x10\x11\x12\x13\x14", 19, False, False, True),
        mkf(0xBADC0FE, b"\x15\x16\x17\x18\x19\x1a\x1b", 19, False, False, False),
        mkf(0xBADC0FE, b"\x1c\x1d\x35\x54", 19, False, True, True),
    ] == list(run(0xBADC0FE, 323219, [mv(bytes(range(0x1E)))], 7))

    crc = pyuavcan.transport.commons.crc.CRC16CCITT()
    crc.add(bytes(range(0x1D)))
    assert crc.value == 0xC46F
    assert [
        mkf(123456, b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e", 19, True, False, True),
        mkf(123456, b"\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\xc4", 19, False, False, False),
        mkf(123456, b"\x6f", 19, False, True, True),
    ] == list(run(123456, 32323219, [mv(bytes(range(0x1D)))], 15))

    crc = pyuavcan.transport.commons.crc.CRC16CCITT()
    crc.add(bytes(range(0x1E)) + b"\x00")
    assert crc.value == 0x32F6
    assert [
        mkf(123456, b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a", 19, True, False, True),
        mkf(123456, b"\x0b\x0c\x0d\x0e\x0f\x10\x11\x12\x13\x14\x15", 19, False, False, False),
        mkf(123456, b"\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x00\x32\xF6", 19, False, True, True),
    ] == list(run(123456, 32323219, [mv(bytes(range(0x1E)))], 11))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import dataclasses
import pyuavcan.transport
import pyuavcan.transport.can


_CANID_EXT_MASK = 2 ** 29 - 1

_BIT_SRV_NOT_MSG = 1 << 25
_BIT_MSG_ANON = 1 << 24
_BIT_SRV_REQ = 1 << 24
_BIT_R23 = 1 << 23
_BIT_MSG_SET_IGNORE = 3 << 21
_BIT_MSG_R7 = 1 << 7


@dataclasses.dataclass(frozen=True)
class CANID:
    PRIORITY_MASK = 7
    NODE_ID_MASK = 127

    priority: pyuavcan.transport.Priority
    source_node_id: typing.Optional[int]  # None if anonymous; may be non-optional in derived classes

    def __post_init__(self) -> None:
        assert isinstance(self.priority, pyuavcan.transport.Priority)

    def compile(self, fragmented_transfer_payload: typing.Iterable[memoryview]) -> int:
        # You might be wondering, why the hell would a CAN ID abstraction depend on the payload of the transfer?
        # This is to accommodate the special case of anonymous message transfers. We need to know the payload to
        # compute the pseudo node ID when emitting anonymous messages. We could use just random numbers from the
        # standard library, but that would make the code hard to test.
        raise NotImplementedError

    @property
    def data_specifier(self) -> pyuavcan.transport.DataSpecifier:
        raise NotImplementedError

    def get_destination_node_id(self) -> typing.Optional[int]:
        """Hides the destination selection logic from users of the abstract type."""
        raise NotImplementedError

    @staticmethod
    def parse(identifier: int) -> typing.Optional[CANID]:
        """
        Attempts to parse the supplied CAN ID value.
        Returns None if the CAN ID is not valid for UAVCAN (different protocol or different version of UAVCAN).
        """
        _validate_unsigned_range(identifier, _CANID_EXT_MASK)
        priority = pyuavcan.transport.Priority(identifier >> 26)
        source_node_id = identifier & CANID.NODE_ID_MASK
        if identifier & _BIT_SRV_NOT_MSG:
            if identifier & _BIT_R23:
                return None  # Wrong protocol
            return ServiceCANID(
                priority=priority,
                service_id=(identifier >> 14) & pyuavcan.transport.ServiceDataSpecifier.SERVICE_ID_MASK,
                request_not_response=identifier & _BIT_SRV_REQ != 0,
                source_node_id=source_node_id,
                destination_node_id=(identifier >> 7) & CANID.NODE_ID_MASK,
            )
        if identifier & (_BIT_R23 | _BIT_MSG_R7):
            return None  # Wrong protocol
        return MessageCANID(
            priority=priority,
            subject_id=(identifier >> 8) & pyuavcan.transport.MessageDataSpecifier.SUBJECT_ID_MASK,
            source_node_id=None if identifier & _BIT_MSG_ANON else source_node_id,
        )


@dataclasses.dataclass(frozen=True)
class MessageCANID(CANID):
    subject_id: int

    def __post_init__(self) -> None:
        super().__post_init__()
        _validate_unsigned_range(int(self.priority), self.PRIORITY_MASK)
        _validate_unsigned_range(self.subject_id, pyuavcan.transport.MessageDataSpecifier.SUBJECT_ID_MASK)
        if self.source_node_id is not None:
            _validate_unsigned_range(self.source_node_id, self.NODE_ID_MASK)

    def compile(self, fragmented_transfer_payload: typing.Iterable[memoryview]) -> int:
        identifier = (int(self.priority) << 26) | _BIT_MSG_SET_IGNORE | (self.subject_id << 8)

        source_node_id = self.source_node_id
        if source_node_id is None:  # Anonymous frame
            # Anonymous transfers cannot be multi-frame, but we have no way of enforcing this here since we don't
            # know what the MTU is. The caller must enforce this instead.
            source_node_id = int(sum(map(sum, fragmented_transfer_payload))) & self.NODE_ID_MASK
            identifier |= _BIT_MSG_ANON

        assert 0 <= source_node_id <= self.NODE_ID_MASK  # Should be valid here already
        identifier |= source_node_id

        assert 0 <= identifier <= _CANID_EXT_MASK
        assert identifier & self.NODE_ID_MASK == source_node_id
        assert (identifier >> 8) & pyuavcan.transport.MessageDataSpecifier.SUBJECT_ID_MASK == self.subject_id
        assert identifier >> 26 == int(self.priority)
        return identifier

    @property
    def data_specifier(self) -> pyuavcan.transport.MessageDataSpecifier:
        return pyuavcan.transport.MessageDataSpecifier(self.subject_id)

    def get_destination_node_id(self) -> typing.Optional[int]:
        return None


@dataclasses.dataclass(frozen=True)
class ServiceCANID(CANID):
    source_node_id: int  # Overrides Optional[int] by covariance (property not writeable)
    destination_node_id: int
    service_id: int
    request_not_response: bool

    def __post_init__(self) -> None:
        super().__post_init__()
        _validate_unsigned_range(int(self.priority), self.PRIORITY_MASK)
        _validate_unsigned_range(self.service_id, pyuavcan.transport.ServiceDataSpecifier.SERVICE_ID_MASK)
        _validate_unsigned_range(self.source_node_id, self.NODE_ID_MASK)
        _validate_unsigned_range(self.destination_node_id, self.NODE_ID_MASK)

        if self.source_node_id == self.destination_node_id:
            raise ValueError(f"Invalid service frame: source node ID == destination node ID == {self.source_node_id}")

    def compile(self, fragmented_transfer_payload: typing.Iterable[memoryview]) -> int:
        del fragmented_transfer_payload
        identifier = (
            (int(self.priority) << 26)
            | _BIT_SRV_NOT_MSG
            | (self.service_id << 14)
            | (self.destination_node_id << 7)
            | self.source_node_id
        )

        if self.request_not_response:
            identifier |= _BIT_SRV_REQ

        assert 0 <= identifier <= _CANID_EXT_MASK
        assert identifier & self.NODE_ID_MASK == self.source_node_id
        assert (identifier >> 14) & 1023 == self.service_id
        assert identifier >> 26 == int(self.priority)
        return identifier

    @property
    def data_specifier(self) -> pyuavcan.transport.ServiceDataSpecifier:
        role_enum = pyuavcan.transport.ServiceDataSpecifier.Role
        role = role_enum.REQUEST if self.request_not_response else role_enum.RESPONSE
        return pyuavcan.transport.ServiceDataSpecifier(self.service_id, role)

    def get_destination_node_id(self) -> typing.Optional[int]:
        return self.destination_node_id


def _validate_unsigned_range(value: int, max_value: int) -> None:
    if not isinstance(value, int) or not (0 <= value <= max_value):
        raise ValueError(f"Value {value} is not in the interval [0, {max_value}]")


def generate_filter_configurations(
    subject_id_list: typing.Iterable[int], local_node_id: typing.Optional[int]
) -> typing.Sequence[pyuavcan.transport.can.media.FilterConfiguration]:
    from .media import FrameFormat, FilterConfiguration

    def ext(idn: int, msk: int) -> FilterConfiguration:
        assert idn <= _CANID_EXT_MASK and msk <= _CANID_EXT_MASK
        return FilterConfiguration(identifier=idn, mask=msk, format=FrameFormat.EXTENDED)

    full: typing.List[FilterConfiguration] = []

    if local_node_id is not None:
        assert local_node_id <= CANID.NODE_ID_MASK
        # If the local node-ID is set, we may receive service requests, so we need to allocate one filter for those.
        full.append(
            ext(
                idn=_BIT_SRV_NOT_MSG | (int(local_node_id) << 7),
                msk=_BIT_SRV_NOT_MSG | _BIT_R23 | (CANID.NODE_ID_MASK << 7),
            )
        )
        # Also, we may need loopback frames for timestamping, so we add a filter for frames where the source node-ID
        # equals ours. Both messages and services!
        full.append(ext(idn=int(local_node_id), msk=_BIT_R23 | CANID.NODE_ID_MASK))
    else:
        # If the local node-ID is not set, we may need to receive loopback frames for sent anonymous transfers.
        # This essentially means that we need to allow ALL anonymous transfers. Those may be only messages, as there
        # is no such thing as anonymous service transfer.
        full.append(ext(idn=_BIT_MSG_ANON, msk=_BIT_SRV_NOT_MSG | _BIT_MSG_ANON | _BIT_R23 | _BIT_MSG_R7))

    # One filter per unique subject-ID. Sorted for testability.
    for sid in sorted(set(subject_id_list)):
        s_mask = pyuavcan.transport.MessageDataSpecifier.SUBJECT_ID_MASK
        assert sid <= s_mask
        full.append(ext(idn=int(sid) << 8, msk=_BIT_SRV_NOT_MSG | _BIT_R23 | (s_mask << 8) | _BIT_MSG_R7))

    return full


def _unittest_can_filter_configuration() -> None:
    from .media import FilterConfiguration, optimize_filter_configurations, FrameFormat

    def ext(idn: int, msk: int) -> FilterConfiguration:
        assert idn <= _CANID_EXT_MASK and msk <= _CANID_EXT_MASK
        return FilterConfiguration(identifier=idn, mask=msk, format=FrameFormat.EXTENDED)

    degenerate = optimize_filter_configurations(generate_filter_configurations([], None), 999)
    assert degenerate == [
        ext(
            idn=0b_000_0_1_0_000000000000000_0_0000000, msk=0b_000_1_1_1_000000000000000_1_0000000  # Anonymous messages
        )
    ]

    no_subjects = optimize_filter_configurations(generate_filter_configurations([], 0b1010101), 999)
    assert no_subjects == [
        ext(idn=0b_000_1_0_0_000000000_1010101_0000000, msk=0b_000_1_0_1_000000000_1111111_0000000),  # Services
        ext(
            idn=0b_000_0_0_0_0000000000000000_1010101,  # Loopback frames (both messages and services)
            msk=0b_000_0_0_1_0000000000000000_1111111,
        ),
    ]

    reference_subject_ids = [
        0b0000000000000000,
        0b0000000000000101,
        0b0000000000001010,
        0b0000000000010101,
        0b0000000000101010,
        0b0000000000101010,  # Duplicate
        0b0000000000101010,  # Triplicate
        0b0000000000101011,  # Similar, Hamming distance 1
    ]

    retained = optimize_filter_configurations(generate_filter_configurations(reference_subject_ids, 0b1010101), 999)
    assert retained == [
        ext(idn=0b_000_1_0_0_000000000_1010101_0000000, msk=0b_000_1_0_1_000000000_1111111_0000000),  # Services
        ext(
            idn=0b_000_0_0_0_0000000000000000_1010101,  # Loopback frames (both messages and services)
            msk=0b_000_0_0_1_0000000000000000_1111111,
        ),
        ext(idn=0b_000_0_0_0_000000000000000_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
        ext(idn=0b_000_0_0_0_000000000000101_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
        ext(idn=0b_000_0_0_0_000000000001010_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
        ext(idn=0b_000_0_0_0_000000000010101_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
        ext(
            idn=0b_000_0_0_0_000000000101010_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000
        ),  # Duplicates removed
        ext(idn=0b_000_0_0_0_000000000101011_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
    ]

    reduced = optimize_filter_configurations(generate_filter_configurations(reference_subject_ids, 0b1010101), 7)
    assert reduced == [
        ext(idn=0b_000_1_0_0_000000000_1010101_0000000, msk=0b_000_1_0_1_000000000_1111111_0000000),  # Services
        ext(
            idn=0b_000_0_0_0_0000000000000000_1010101,  # Loopback frames (both messages and services)
            msk=0b_000_0_0_1_0000000000000000_1111111,
        ),
        ext(idn=0b_000_0_0_0_000000000000000_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
        ext(idn=0b_000_0_0_0_000000000000101_0_0000000, msk=0b_000_1_0_1_001111111101111_1_0000000),  # Merged with 6th
        ext(idn=0b_000_0_0_0_000000000001010_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
        # This one removed, merged with 4th
        ext(
            idn=0b_000_0_0_0_000000000101010_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000
        ),  # Duplicates removed
        ext(idn=0b_000_0_0_0_000000000101011_0_0000000, msk=0b_000_1_0_1_001111111111111_1_0000000),
    ]
    print([str(r) for r in reduced])

    reduced = optimize_filter_configurations(generate_filter_configurations(reference_subject_ids, 0b1010101), 3)
    assert reduced == [
        ext(idn=0b_000_1_0_0_000000000_1010101_0000000, msk=0b_000_1_0_1_000000000_1111111_0000000),  # Services
        ext(
            idn=0b_000_0_0_0_0000000000000000_1010101,  # Loopback frames (both messages and services)
            msk=0b_000_0_0_1_0000000000000000_1111111,
        ),
        ext(idn=0b_000_0_0_0_000000000000000_0_0000000, msk=0b_000_1_0_1_001111111000000_1_0000000),
    ]
    print([str(r) for r in reduced])

    reduced = optimize_filter_configurations(generate_filter_configurations(reference_subject_ids, 0b1010101), 1)
    assert reduced == [
        ext(
            idn=0b_000_0_0_0_000000000_0000000_0000000, msk=0b_000_0_0_1_000000000_0000000_0000000
        ),  # Degenerates to checking only the reserved bits
    ]
    print([str(r) for r in reduced])


def _unittest_can_identifier_parse() -> None:
    from pytest import raises
    from pyuavcan.transport import Priority, MessageDataSpecifier, ServiceDataSpecifier

    with raises(ValueError):
        CANID.parse(_CANID_EXT_MASK + 1)

    with raises(ValueError):
        MessageCANID(Priority.HIGH, None, 2 ** 15)

    with raises(ValueError):
        MessageCANID(Priority.HIGH, 128, 123)

    with raises(ValueError):
        MessageCANID(Priority.HIGH, 123, -1)

    with raises(ValueError):
        MessageCANID(Priority.HIGH, -1, 123)

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, -1, 123, 123, True)

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, 128, 123, 123, True)

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, 123, -1, 123, True)

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, 123, 128, 123, True)

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, 123, 123, -1, True)

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, 123, 123, 512, True)

    with raises(ValueError):
        # noinspection PyTypeChecker
        ServiceCANID(Priority.HIGH, None, 123, 512, True)  # type: ignore

    with raises(ValueError):
        ServiceCANID(Priority.HIGH, 123, 123, 42, True)  # Same source and destination

    assert CANID.parse(0b_010_0_0_0110100100101001_1_1111011) is None
    reference_message = MessageCANID(Priority.FAST, 123, 2345)
    assert CANID.parse(0b_010_0_0_0110100100101001_0_1111011) == reference_message
    assert CANID.parse(0b_010_0_0_0100100100101001_0_1111011) == reference_message
    assert CANID.parse(0b_010_0_0_0010100100101001_0_1111011) == reference_message
    assert CANID.parse(0b_010_0_0_0000100100101001_0_1111011) == reference_message
    assert 0b_010_0_0_0110100100101001_0_1111011 == reference_message.compile([])
    assert reference_message.data_specifier == MessageDataSpecifier(2345)

    assert CANID.parse(0b_010_0_1_0111000011100001_1_1111111) is None
    reference_message = MessageCANID(Priority.FAST, None, 4321)
    assert CANID.parse(0b_010_0_1_0111000011100001_0_1111111) == reference_message
    assert CANID.parse(0b_010_0_1_0101000011100001_0_1111111) == reference_message
    assert CANID.parse(0b_010_0_1_0011000011100001_0_1111111) == reference_message
    assert CANID.parse(0b_010_0_1_0001000011100001_0_1111111) == reference_message
    assert 0b_010_0_1_0111000011100001_0_1111111 == reference_message.compile([memoryview(bytes([100, 27]))])
    assert reference_message.data_specifier == MessageDataSpecifier(4321)

    reference_service = ServiceCANID(Priority.OPTIONAL, 123, 42, 300, True)
    reference_service_id = 0b_111_1_1_0100101100_0101010_1111011
    assert CANID.parse(reference_service_id) == reference_service
    assert reference_service_id == reference_service.compile([])
    assert reference_service.data_specifier == ServiceDataSpecifier(300, ServiceDataSpecifier.Role.REQUEST)

    reference_service = ServiceCANID(Priority.OPTIONAL, 42, 123, 255, False)
    reference_service_id = 0b_111_1_0_0011111111_1111011_0101010
    assert CANID.parse(reference_service_id) == reference_service
    assert reference_service_id == reference_service.compile([])
    assert reference_service.data_specifier == ServiceDataSpecifier(255, ServiceDataSpecifier.Role.RESPONSE)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import socket
import logging
import ipaddress
import pyuavcan.util
import pyuavcan.transport
from ._endpoint_mapping import IPAddress
from ._link_layer import LinkLayerCapture


_logger = logging.getLogger(__name__)


class SocketFactory(abc.ABC):
    """
    The factory encapsulates the mapping logic between data specifiers and UDP endpoints.
    Additionally, it also provides an abstract interface for constructing IP-version-specific sniffers.

    May be related:

    - https://stackoverflow.com/a/26988214/1007777
    - https://stackoverflow.com/a/14388707/1007777
    - https://tldp.org/HOWTO/Multicast-HOWTO-6.html
    - https://habr.com/ru/post/141021/
    - https://habr.com/ru/company/cbs/blog/309486/
    - https://stackoverflow.com/a/58118503/1007777
    - http://www.enderunix.org/docs/en/rawipspoof/
    - https://docs.oracle.com/cd/E19683-01/816-5042/sockets-5/index.html
    """

    MULTICAST_TTL = 16
    """
    RFC 1112 dictates that the default TTL for multicast sockets is 1.
    This is not acceptable so we use a larger default.
    """

    @staticmethod
    def new(local_ip_address: IPAddress) -> SocketFactory:
        """
        Use this factory factory to create new instances.
        """
        if isinstance(local_ip_address, ipaddress.IPv4Address):
            from ._v4 import IPv4SocketFactory

            return IPv4SocketFactory(local_ip_address)

        if isinstance(local_ip_address, ipaddress.IPv6Address):
            raise NotImplementedError("Sorry, IPv6 is not yet supported by this implementation.")

        raise TypeError(f"Invalid local IP address: {local_ip_address!r}")  # pragma: no cover

    @property
    @abc.abstractmethod
    def max_nodes(self) -> int:
        """
        The maximum number of nodes per subnet may be a function of the protocol version.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def local_ip_address(self) -> IPAddress:
        raise NotImplementedError

    @abc.abstractmethod
    def make_output_socket(
        self, remote_node_id: typing.Optional[int], data_specifier: pyuavcan.transport.DataSpecifier
    ) -> socket.socket:
        """
        Make a new non-blocking output socket connected to the appropriate endpoint
        (unicast for service data specifiers, multicast for message data specifiers).
        The socket will be bound to an ephemeral port at the configured local network address.

        The required options will be set up as needed automatically.
        Timestamping will need to be enabled separately if needed.

        WARNING: on Windows, multicast output sockets have a weird corner case.
        If the output interface is set to the loopback adapter and there are no registered listeners for the specified
        multicast group, an attempt to send data to that group will fail with a "network unreachable" error.
        Here is an example::

            import socket, asyncio
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.bind(('127.1.2.3', 0))
            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton('127.1.2.3'))
            s.sendto(b'\xaa\xbb\xcc', ('127.5.5.5', 1234))          # Success
            s.sendto(b'\xaa\xbb\xcc', ('239.1.2.3', 1234))          # OSError
            # OSError: [WinError 10051] A socket operation was attempted to an unreachable network
            loop = asyncio.get_event_loop()
            loop.run_until_complete(loop.sock_sendall(s, b'abc'))   # OSError
            # OSError: [WinError 1231] The network location cannot be reached
        """
        raise NotImplementedError

    @abc.abstractmethod
    def make_input_socket(self, data_specifier: pyuavcan.transport.DataSpecifier) -> socket.socket:
        r"""
        Makes a new non-blocking input socket bound to the correct endpoint
        (unicast for service data specifiers, multicast for message data specifiers).

        The required socket options will be set up as needed automatically;
        specifically, ``SO_REUSEADDR``, ``SO_REUSEPORT`` (if available), maybe others as needed.
        Timestamping will need to be enabled separately if needed.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def make_sniffer(self, handler: typing.Callable[[LinkLayerCapture], None]) -> Sniffer:
        """
        Launch a new network sniffer based on a raw socket (usually this requires special permissions).
        The sniffer will run in a separate thread, invoking the handler *directly from the worker thread*
        whenever a UDP packet from the specified subnet is received.

        Packets whose origin does not belong to the current UAVCAN/UDP subnet are dropped (not reported).
        This is critical because there may be multiple UAVCAN/UDP transport networks running on the same
        physical IP network, which may also be shared with other protocols.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, local_ip_address=str(self.local_ip_address))


class Sniffer(abc.ABC):
    """
    Network sniffer is responsible for managing the raw socket and parsing and filtering the raw IP packets.
    """

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
The module is always importable but is functional only on GNU/Linux.

For testing or experimentation on a local machine it is often convenient to use a virtual CAN bus instead of a real one.
Using SocketCAN, one can set up a virtual CAN bus interface as follows::

    modprobe can
    modprobe can_raw
    modprobe vcan
    ip link add dev vcan0 type vcan
    ip link set vcan0 mtu 72         # Enable CAN FD by configuring the MTU of 64+8
    ip link set up vcan0

Where ``vcan0`` can be replaced with any other valid interface name.
Please read the SocketCAN documentation for more information.
"""

from sys import platform as _platform

if _platform == "linux":
    from ._socketcan import SocketCANMedia as SocketCANMedia



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
The util package contains various entities that are commonly useful in PyUAVCAN-based applications.
"""

from ._broadcast import broadcast as broadcast

from ._introspect import import_submodules as import_submodules
from ._introspect import iter_descendants as iter_descendants

from ._mark_last import mark_last as mark_last

from ._repr import repr_attributes as repr_attributes
from ._repr import repr_attributes_noexcept as repr_attributes_noexcept



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._base import Deduplicator as Deduplicator

from ._monotonic import MonotonicDeduplicator as MonotonicDeduplicator

from ._cyclic import CyclicDeduplicator as CyclicDeduplicator



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import asyncio
import logging
import dataclasses
import pyuavcan.dsdl
import pyuavcan.transport
from ._base import ServiceClass, ServicePort, PortFinalizer, OutgoingTransferIDCounter, Closable
from ._base import DEFAULT_PRIORITY, DEFAULT_SERVICE_REQUEST_TIMEOUT
from ._error import PortClosedError, RequestTransferIDVariabilityExhaustedError


# Shouldn't be too large as this value defines how quickly the task will detect that the underlying transport is closed.
_RECEIVE_TIMEOUT = 1


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ClientStatistics:
    """
    The counters are maintained at the hidden client instance which is not accessible to the user.
    As such, clients with the same session specifier will share the same set of statistical counters.
    """

    request_transport_session: pyuavcan.transport.SessionStatistics
    response_transport_session: pyuavcan.transport.SessionStatistics
    sent_requests: int
    deserialization_failures: int  #: Response transfers that could not be deserialized into a response object.
    unexpected_responses: int  #: Response transfers that could not be matched with a request state.


class Client(ServicePort[ServiceClass]):
    """
    A task should request its own client instance from the presentation layer controller.
    Do not share the same client instance across different tasks. This class implements the RAII pattern.

    Implementation info: all client instances sharing the same session specifier also share the same
    underlying implementation object containing the transport sessions which is reference counted and
    destroyed automatically when the last client instance is closed;
    the user code cannot access it and generally shouldn't care.
    None of the settings of a client instance, such as timeout or priority, can affect other client instances;
    this does not apply to the transfer-ID counter objects though because they are transport-layer entities
    and therefore are shared per session specifier.

    .. note::
        Normally we should use correct generic types ``ServiceClass.Request`` and ``ServiceClass.Response`` in the API;
        however, MyPy does not support that yet. Please find the context at
        https://github.com/python/mypy/issues/7121 (please upvote!) and https://github.com/UAVCAN/pyuavcan/issues/61.
        We use a tentative workaround for now to silence bogus type errors. When the missing logic is implemented
        in MyPy, this should be switched back to proper implementation.
    """

    def __init__(self, impl: ClientImpl[ServiceClass], loop: asyncio.AbstractEventLoop):
        """
        Do not call this directly! Use :meth:`Presentation.make_client`.
        """
        assert not impl.is_closed, "Internal logic error"
        self._maybe_impl: typing.Optional[ClientImpl[ServiceClass]] = impl
        impl.register_proxy()  # Register ASAP to ensure correct finalization.

        self._loop = loop
        self._dtype = impl.dtype  # Permit usage after close()
        self._input_transport_session = impl.input_transport_session  # Same
        self._output_transport_session = impl.output_transport_session  # Same
        self._transfer_id_counter = impl.transfer_id_counter  # Same
        self._response_timeout = DEFAULT_SERVICE_REQUEST_TIMEOUT
        self._priority = DEFAULT_PRIORITY

    async def call(
        self, request: pyuavcan.dsdl.CompositeObject
    ) -> typing.Optional[typing.Tuple[pyuavcan.dsdl.CompositeObject, pyuavcan.transport.TransferFrom]]:
        """
        Sends the request to the remote server using the pre-configured priority and response timeout parameters.
        Returns the response along with its transfer info in the case of successful completion.
        If the server did not provide a valid response on time, returns None.

        On certain feature-limited transfers (such as CAN) the call may raise
        :class:`pyuavcan.presentation.RequestTransferIDVariabilityExhaustedError`
        if there are too many concurrent requests.
        """
        if self._maybe_impl is None:
            raise PortClosedError(repr(self))
        return await self._maybe_impl.call(
            request=request, priority=self._priority, response_timeout=self._response_timeout
        )

    @property
    def response_timeout(self) -> float:
        """
        The response timeout value used for requests emitted via this proxy instance.
        This parameter is configured separately per proxy instance; i.e., it is not shared across different client
        instances under the same session specifier, so that, for example, different tasks invoking the same service
        on the same server node can have different timeout settings.
        The same value is also used as send timeout for the underlying call to
        :meth:`pyuavcan.transport.OutputSession.send`.
        The default value is set according to the recommendations provided in the Specification,
        which is :data:`DEFAULT_SERVICE_REQUEST_TIMEOUT`.
        """
        return self._response_timeout

    @response_timeout.setter
    def response_timeout(self, value: float) -> None:
        value = float(value)
        if 0 < value < float("+inf"):
            self._response_timeout = float(value)
        else:
            raise ValueError(f"Invalid response timeout value: {value}")

    @property
    def priority(self) -> pyuavcan.transport.Priority:
        """
        The priority level used for requests emitted via this proxy instance.
        This parameter is configured separately per proxy instance; i.e., it is not shared across different client
        instances under the same session specifier.
        """
        return self._priority

    @priority.setter
    def priority(self, value: pyuavcan.transport.Priority) -> None:
        self._priority = pyuavcan.transport.Priority(value)

    @property
    def dtype(self) -> typing.Type[ServiceClass]:
        return self._dtype

    @property
    def transfer_id_counter(self) -> OutgoingTransferIDCounter:
        """
        Allows the caller to reach the transfer-ID counter instance.
        The instance is shared for clients under the same session.
        I.e., if there are two clients that use the same service-ID and same server node-ID,
        they will share the same transfer-ID counter.
        """
        return self._transfer_id_counter

    @property
    def input_transport_session(self) -> pyuavcan.transport.InputSession:
        return self._input_transport_session

    @property
    def output_transport_session(self) -> pyuavcan.transport.OutputSession:
        """
        The transport session used for request transfers.
        """
        return self._output_transport_session

    def sample_statistics(self) -> ClientStatistics:
        """
        The statistics are counted at the hidden implementation instance.
        Clients that use the same session specifier will have the same set of statistical counters.
        """
        if self._maybe_impl is None:
            raise PortClosedError(repr(self))
        return ClientStatistics(
            request_transport_session=self.output_transport_session.sample_statistics(),
            response_transport_session=self.input_transport_session.sample_statistics(),
            sent_requests=self._maybe_impl.sent_request_count,
            deserialization_failures=self._maybe_impl.deserialization_failure_count,
            unexpected_responses=self._maybe_impl.unexpected_response_count,
        )

    def close(self) -> None:
        impl, self._maybe_impl = self._maybe_impl, None
        if impl is not None:
            impl.remove_proxy()

    def __del__(self) -> None:
        if self._maybe_impl is not None:
            # https://docs.python.org/3/reference/datamodel.html#object.__del__
            # DO NOT invoke logging from the finalizer because it may resurrect the object!
            # Once it is resurrected, we may run into resource management issue if __del__() is invoked again.
            # Whether it is invoked the second time is an implementation detail.
            # If it is invoked again, then we may terminate the client implementation prematurely, leaving existing
            # client proxy instances with a dead reference to a finalized implementation.
            # RAII is difficult in Python. Maybe we should require the user to manage resources manually?
            self._maybe_impl.remove_proxy()
            self._maybe_impl = None


class ClientImpl(Closable, typing.Generic[ServiceClass]):
    """
    The client implementation. There is at most one such implementation per session specifier. It may be shared
    across multiple users with the help of the proxy class. When the last proxy is closed or garbage collected,
    the implementation will also be closed and removed. This is not a part of the library API.
    """

    def __init__(
        self,
        dtype: typing.Type[ServiceClass],
        input_transport_session: pyuavcan.transport.InputSession,
        output_transport_session: pyuavcan.transport.OutputSession,
        transfer_id_counter: OutgoingTransferIDCounter,
        transfer_id_modulo_factory: typing.Callable[[], int],
        finalizer: PortFinalizer,
        loop: asyncio.AbstractEventLoop,
    ):
        self.dtype = dtype
        self.input_transport_session = input_transport_session
        self.output_transport_session = output_transport_session

        self.sent_request_count = 0
        self.unsent_request_count = 0
        self.deserialization_failure_count = 0
        self.unexpected_response_count = 0

        self.transfer_id_counter = transfer_id_counter
        # The transfer ID modulo may change if the transport is reconfigured at runtime. This is certainly not a
        # common use case, but it makes sense supporting it in this library since it's supposed to be usable with
        # diagnostic and inspection tools.
        self._transfer_id_modulo_factory = transfer_id_modulo_factory
        self._maybe_finalizer: typing.Optional[PortFinalizer] = finalizer
        self._loop = loop

        self._lock = asyncio.Lock()
        self._proxy_count = 0
        self._response_futures_by_transfer_id: typing.Dict[
            int, asyncio.Future[typing.Tuple[pyuavcan.dsdl.CompositeObject, pyuavcan.transport.TransferFrom]]
        ] = {}

        self._task = loop.create_task(self._task_function())

    @property
    def is_closed(self) -> bool:
        return self._maybe_finalizer is None

    async def call(
        self, request: pyuavcan.dsdl.CompositeObject, priority: pyuavcan.transport.Priority, response_timeout: float
    ) -> typing.Optional[typing.Tuple[pyuavcan.dsdl.CompositeObject, pyuavcan.transport.TransferFrom]]:
        async with self._lock:
            if self.is_closed:
                raise PortClosedError(repr(self))

            # We have to compute the modulus here manually instead of just letting the transport do that because
            # the response will use the modulus instead of the full TID and we have to match it with the request.
            transfer_id = self.transfer_id_counter.get_then_increment() % self._transfer_id_modulo_factory()
            if transfer_id in self._response_futures_by_transfer_id:
                raise RequestTransferIDVariabilityExhaustedError(repr(self))

            try:
                future = self._loop.create_future()
                self._response_futures_by_transfer_id[transfer_id] = future
                # The lock is still taken, this is intentional. Serialize access to the transport.
                send_result = await self._do_send(
                    request=request,
                    transfer_id=transfer_id,
                    priority=priority,
                    monotonic_deadline=self._loop.time() + response_timeout,
                )
            except BaseException:
                self._forget_future(transfer_id)
                raise

        # Wait for the response with the lock released.
        # We have to make sure that no matter what happens, we remove the future from the table upon exit;
        # otherwise the user will get a false exception when the same transfer ID is reused (which only happens
        # with some low-capability transports such as CAN bus though).
        try:
            if send_result:
                self.sent_request_count += 1
                response, transfer = await asyncio.wait_for(future, timeout=response_timeout)
                assert isinstance(response, self.dtype.Response)
                assert isinstance(transfer, pyuavcan.transport.TransferFrom)
                return response, transfer
            self.unsent_request_count += 1
            return None
        except asyncio.TimeoutError:
            return None
        finally:
            self._forget_future(transfer_id)

    def register_proxy(self) -> None:  # Proxy (de-)registration is always possible even if closed.
        assert not self.is_closed, "Internal logic error: cannot register a new proxy on a closed instance"
        assert self._proxy_count >= 0
        self._proxy_count += 1
        _logger.debug("%s got a new proxy, new count %s", self, self._proxy_count)

    def remove_proxy(self) -> None:
        self._proxy_count -= 1
        _logger.debug("%s has lost a proxy, new count %s", self, self._proxy_count)
        assert self._proxy_count >= 0
        if self._proxy_count <= 0:
            self.close()  # RAII auto-close

    @property
    def proxy_count(self) -> int:
        """Testing facilitation."""
        assert self._proxy_count >= 0
        return self._proxy_count

    def close(self) -> None:
        try:
            self._task.cancel()
        except Exception as ex:
            _logger.debug("Could not cancel the task %r: %s", self._task, ex, exc_info=True)
        self._finalize()

    async def _do_send(
        self,
        request: pyuavcan.dsdl.CompositeObject,
        transfer_id: int,
        priority: pyuavcan.transport.Priority,
        monotonic_deadline: float,
    ) -> bool:
        if not isinstance(request, self.dtype.Request):
            raise TypeError(
                f"Invalid request object: expected an instance of {self.dtype.Request}, "
                f"got {type(request)} instead."
            )

        timestamp = pyuavcan.transport.Timestamp.now()
        fragmented_payload = list(pyuavcan.dsdl.serialize(request))
        transfer = pyuavcan.transport.Transfer(
            timestamp=timestamp, priority=priority, transfer_id=transfer_id, fragmented_payload=fragmented_payload
        )
        return await self.output_transport_session.send(transfer, monotonic_deadline)

    async def _task_function(self) -> None:
        exception: typing.Optional[Exception] = None
        try:
            while not self.is_closed:
                transfer = await self.input_transport_session.receive(self._loop.time() + _RECEIVE_TIMEOUT)
                if transfer is None:
                    continue

                response = pyuavcan.dsdl.deserialize(self.dtype.Response, transfer.fragmented_payload)
                if response is None:
                    self.deserialization_failure_count += 1
                    continue

                try:
                    fut = self._response_futures_by_transfer_id.pop(transfer.transfer_id)
                except LookupError:
                    _logger.info(
                        "Unexpected response %s with transfer %s; TID values of pending requests: %r",
                        response,
                        transfer,
                        list(self._response_futures_by_transfer_id.keys()),
                    )
                    self.unexpected_response_count += 1
                else:
                    fut.set_result((response, transfer))
        except asyncio.CancelledError:
            _logger.debug("Cancelling the task of %s", self)
        except Exception as ex:
            exception = ex
            # Do not use f-string because it can throw, unlike the built-in formatting facility of the logger
            _logger.exception("Fatal error in the task of %s: %s", self, ex)
        finally:
            self._finalize(exception)
            assert self.is_closed

    def _forget_future(self, transfer_id: int) -> None:
        try:
            del self._response_futures_by_transfer_id[transfer_id]
        except LookupError:
            pass

    def _finalize(self, exception: typing.Optional[Exception] = None) -> None:
        exception = exception if exception is not None else PortClosedError(repr(self))
        try:
            if self._maybe_finalizer is not None:
                self._maybe_finalizer([self.input_transport_session, self.output_transport_session])
                self._maybe_finalizer = None
        except Exception as ex:
            _logger.exception("%s failed to finalize: %s", self, ex)
        for fut in self._response_futures_by_transfer_id.values():
            try:
                fut.set_exception(exception)
            except asyncio.InvalidStateError:
                pass

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes_noexcept(
            self,
            dtype=str(pyuavcan.dsdl.get_model(self.dtype)),
            input_transport_session=self.input_transport_session,
            output_transport_session=self.output_transport_session,
            proxy_count=self._proxy_count,
        )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import copy
import typing
import asyncio
import logging
import dataclasses
import pyuavcan.transport
from pyuavcan.transport import Timestamp
from .media import Media, Envelope, optimize_filter_configurations, FilterConfiguration, FrameFormat
from ._session import CANInputSession, CANOutputSession, SendTransaction
from ._session import BroadcastCANOutputSession, UnicastCANOutputSession
from ._frame import UAVCANFrame, TRANSFER_ID_MODULO
from ._identifier import CANID, generate_filter_configurations
from ._input_dispatch_table import InputDispatchTable
from ._tracer import CANTracer, CANCapture


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class CANTransportStatistics(pyuavcan.transport.TransportStatistics):
    """
    The following invariants apply::

        out_frames >= out_frames_loopback
        in_frames >= in_frames_uavcan >= in_frames_uavcan_accepted
        out_frames_loopback >= in_frames_loopback
    """

    in_frames: int = 0  #: Number of genuine frames received from the bus (loopback not included).
    in_frames_uavcan: int = 0  #: Subset of the above that happen to be valid UAVCAN frames.
    in_frames_uavcan_accepted: int = 0  #: Subset of the above that are useful for the local application.
    in_frames_loopback: int = 0  #: Number of loopback frames received from the media instance (not bus).
    in_frames_errored: int = 0  #: How many frames of any kind could not be successfully processed.

    out_frames: int = 0  #: Number of frames sent to the media instance successfully.
    out_frames_timeout: int = 0  #: Number of frames that were supposed to be sent but timed out.
    out_frames_loopback: int = 0  #: Number of sent frames that we requested loopback for.

    @property
    def media_acceptance_filtering_efficiency(self) -> float:
        """
        An efficiency metric for the acceptance filtering implemented in the media instance.
        The value of 1.0 (100%) indicates perfect filtering, where the media can sort out relevant frames from
        irrelevant ones completely autonomously. The value of 0 indicates that none of the frames passed over
        from the media instance are useful for the application (all ignored).
        """
        return (self.in_frames_uavcan_accepted / self.in_frames) if self.in_frames > 0 else 1.0

    @property
    def lost_loopback_frames(self) -> int:
        """
        The number of loopback frames that have been requested but never returned. Normally the value should be zero.
        The value may transiently increase to small values if the counters happened to be sampled while the loopback
        frames reside in the transmission queue of the CAN controller awaiting being processed. If the value remains
        positive for long periods of time, the media driver is probably misbehaving.
        A negative value means that the media instance is sending more loopback frames than requested (bad).
        """
        return self.out_frames_loopback - self.in_frames_loopback


class CANTransport(pyuavcan.transport.Transport):
    """
    The standard UAVCAN/CAN transport implementation as defined in the UAVCAN specification.
    Please read the module documentation for details.
    """

    TRANSFER_ID_MODULO = TRANSFER_ID_MODULO

    def __init__(
        self,
        media: Media,
        local_node_id: typing.Optional[int],
        *,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        :param media:         The media implementation.
        :param local_node_id: The node-ID to use. Can't be changed. None means anonymous (useful for PnP allocation).
        :param loop:          The event loop to use. Defaults to :func:`asyncio.get_event_loop`.
        """
        self._maybe_media: typing.Optional[Media] = media
        self._local_node_id = int(local_node_id) if local_node_id is not None else None
        self._media_lock = asyncio.Lock()
        self._loop = loop if loop is not None else asyncio.get_event_loop()

        # Lookup performance for the output registry is not important because it's only used for loopback frames.
        # Hence we don't trade-off memory for speed here.
        self._output_registry: typing.Dict[pyuavcan.transport.OutputSessionSpecifier, CANOutputSession] = {}

        # Input lookup must be fast, so we use constant-complexity static lookup table.
        self._input_dispatch_table = InputDispatchTable()

        self._last_filter_configuration_set: typing.Optional[typing.Sequence[FilterConfiguration]] = None

        self._capture_handlers: typing.List[pyuavcan.transport.CaptureCallback] = []

        self._frame_stats = CANTransportStatistics()

        if self._local_node_id is not None and not 0 <= self._local_node_id <= CANID.NODE_ID_MASK:
            raise ValueError(f"Invalid node ID for CAN: {self._local_node_id}")

        if media.mtu not in Media.VALID_MTU_SET:
            raise pyuavcan.transport.InvalidMediaConfigurationError(
                f"The MTU value {media.mtu} is not a member of {Media.VALID_MTU_SET}"
            )
        self._mtu = media.mtu - 1
        assert self._mtu > 0

        if media.number_of_acceptance_filters < 1:
            raise pyuavcan.transport.InvalidMediaConfigurationError(
                f"The number of acceptance filters is too low: {media.number_of_acceptance_filters}"
            )

        if media.loop is not self._loop:
            raise pyuavcan.transport.InvalidMediaConfigurationError(
                f"The media instance cannot use a different event loop: {media.loop} is not {self._loop}"
            )

        media.start(self._on_frames_received, no_automatic_retransmission=self._local_node_id is None)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def protocol_parameters(self) -> pyuavcan.transport.ProtocolParameters:
        return pyuavcan.transport.ProtocolParameters(
            transfer_id_modulo=TRANSFER_ID_MODULO,
            max_nodes=CANID.NODE_ID_MASK + 1,
            mtu=self._mtu,
        )

    @property
    def local_node_id(self) -> typing.Optional[int]:
        """
        If the local node-ID is not assigned, automatic retransmission in the media implementation is disabled to
        facilitate plug-and-play node-ID allocation.
        """
        return self._local_node_id

    @property
    def input_sessions(self) -> typing.Sequence[CANInputSession]:
        return list(self._input_dispatch_table.items)

    @property
    def output_sessions(self) -> typing.Sequence[CANOutputSession]:
        return list(self._output_registry.values())

    def close(self) -> None:
        for s in (*self.input_sessions, *self.output_sessions):
            try:
                s.close()
            except Exception as ex:
                _logger.exception("%s: Failed to close session %r: %s", self, s, ex)

        media, self._maybe_media = self._maybe_media, None
        if media is not None:  # Double-close is NOT an error!
            media.close()

    def sample_statistics(self) -> CANTransportStatistics:
        return copy.copy(self._frame_stats)

    def get_input_session(
        self, specifier: pyuavcan.transport.InputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> CANInputSession:
        """
        See the base class docs for background.
        Whenever an input session is created or destroyed, the hardware acceptance filters are reconfigured
        automatically; computation of a new configuration and its deployment on the CAN controller may be slow.
        """
        if self._maybe_media is None:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

        def finalizer() -> None:
            self._input_dispatch_table.remove(specifier)
            self._reconfigure_acceptance_filters()

        session = self._input_dispatch_table.get(specifier)
        if session is None:
            session = CANInputSession(
                specifier=specifier, payload_metadata=payload_metadata, loop=self._loop, finalizer=finalizer
            )
            self._input_dispatch_table.add(session)
            self._reconfigure_acceptance_filters()
        return session

    def get_output_session(
        self, specifier: pyuavcan.transport.OutputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> CANOutputSession:
        if self._maybe_media is None:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

        try:
            out = self._output_registry[specifier]
            assert out.specifier == specifier
            assert (specifier.remote_node_id is None) == isinstance(out, BroadcastCANOutputSession)
            return out
        except KeyError:
            pass

        def finalizer() -> None:
            self._output_registry.pop(specifier)

        if specifier.is_broadcast:
            session: CANOutputSession = BroadcastCANOutputSession(
                specifier=specifier,
                payload_metadata=payload_metadata,
                transport=self,
                send_handler=self._do_send,
                finalizer=finalizer,
            )
        else:
            session = UnicastCANOutputSession(
                specifier=specifier,
                payload_metadata=payload_metadata,
                transport=self,
                send_handler=self._do_send,
                finalizer=finalizer,
            )

        self._output_registry[specifier] = session
        if not self._last_filter_configuration_set:
            # It is necessary to reconfigure the filters at least once to ensure that we are able to receive
            # loopback frames even if there are no input sessions in use.
            self._reconfigure_acceptance_filters()
        return session

    def begin_capture(self, handler: pyuavcan.transport.CaptureCallback) -> None:
        """
        Capture is implemented by reconfiguring the acceptance filter to accept everything
        and forcing loopback for every outgoing frame.
        Forced loopback ensures that transmitted frames are timestamped very accurately.
        Captured frames are encapsulated inside :class:`pyuavcan.transport.can.CANCapture`.
        """
        self._capture_handlers.append(handler)
        self._reconfigure_acceptance_filters()

    @property
    def capture_active(self) -> bool:
        return len(self._capture_handlers) > 0

    @staticmethod
    def make_tracer() -> CANTracer:
        """
        See :class:`CANTracer`.
        """
        return CANTracer()

    async def spoof(self, transfer: pyuavcan.transport.AlienTransfer, monotonic_deadline: float) -> bool:
        """
        Spoofing over the CAN transport is trivial and it does not involve reconfiguration of the media layer.
        It can be invoked at no cost at any time (unlike, say, UAVCAN/UDP).
        See the overridden method :meth:`pyuavcan.transport.Transport.spoof` for details.
        """
        from ._session import serialize_transfer
        from ._identifier import MessageCANID, ServiceCANID

        ss = transfer.metadata.session_specifier
        src, dst = ss.source_node_id, ss.destination_node_id
        can_id: CANID
        if isinstance(ss.data_specifier, pyuavcan.transport.MessageDataSpecifier):
            if dst is not None:
                raise pyuavcan.transport.UnsupportedSessionConfigurationError(
                    f"Unicast message transfers are not allowed. Spoof metadata: {transfer.metadata}"
                )
            can_id = MessageCANID(
                priority=transfer.metadata.priority,
                source_node_id=src,
                subject_id=ss.data_specifier.subject_id,
            )
        elif isinstance(ss.data_specifier, pyuavcan.transport.ServiceDataSpecifier):
            if src is None or dst is None:
                raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                    f"Anonymous nodes cannot participate in service calls. Spoof metadata: {transfer.metadata}"
                )
            can_id = ServiceCANID(
                priority=transfer.metadata.priority,
                source_node_id=src,
                destination_node_id=dst,
                service_id=ss.data_specifier.service_id,
                request_not_response=ss.data_specifier.role == pyuavcan.transport.ServiceDataSpecifier.Role.REQUEST,
            )
        else:
            assert False

        frames = list(
            serialize_transfer(
                compiled_identifier=can_id.compile(transfer.fragmented_payload),
                transfer_id=transfer.metadata.transfer_id % TRANSFER_ID_MODULO,
                fragmented_payload=transfer.fragmented_payload,
                max_frame_payload_bytes=self.protocol_parameters.mtu,
            )
        )
        if len(frames) > 1 and src is None:
            raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                f"Anonymous nodes cannot emit multi-frame transfers. Spoof metadata: {transfer.metadata}"
            )
        transaction = SendTransaction(frames, loopback_first=False, monotonic_deadline=monotonic_deadline)
        return await self._do_send(transaction)

    async def _do_send(self, t: SendTransaction) -> bool:
        """
        All frames shall share the same CAN ID value.
        """
        force_loopback = bool(self._capture_handlers)
        async with self._media_lock:
            if self._maybe_media is None:
                raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

            if _logger.isEnabledFor(logging.DEBUG):
                timeout = t.monotonic_deadline - self._loop.time()
                _logger.debug(
                    "%s: Sending %d frames; 1st loopback: %s; deadline in %.3f s:\n%s",
                    self,
                    len(t.frames),
                    t.loopback_first,
                    timeout,
                    "\n".join(map(str, t.frames)),
                )

            num_sent = await self._maybe_media.send(
                (
                    Envelope(
                        frame=x.compile(),
                        loopback=((idx == 0 and t.loopback_first) or force_loopback),
                    )
                    for idx, x in enumerate(t.frames)
                ),
                t.monotonic_deadline,
            )
            assert 0 <= num_sent <= len(t.frames), "Media sub-layer API contract violation"
            sent_frames, unsent_frames = t.frames[:num_sent], t.frames[num_sent:]

            self._frame_stats.out_frames += len(sent_frames)
            self._frame_stats.out_frames_timeout += len(unsent_frames)
            self._frame_stats.out_frames_loopback += 1 if t.loopback_first else 0

        if unsent_frames:
            can_id_int_set = set(f.identifier for f in unsent_frames)
            assert len(can_id_int_set) == 1, "CAN transport layer internal contract violation"
            (can_id_int,) = can_id_int_set
            _logger.info(
                "%s: %d frames of %d total with CAN ID 0x%08x could not be sent before the deadline",
                self,
                len(unsent_frames),
                num_sent,
                can_id_int,
            )

        return not unsent_frames

    def _on_frames_received(self, frames: typing.Sequence[typing.Tuple[Timestamp, Envelope]]) -> None:
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("%s: Parsing received CAN frames:\n%s", self, "\n".join(f"{t} {e}" for t, e in frames))

        for timestamp, envelope in frames:
            try:
                if envelope.loopback:
                    self._frame_stats.in_frames_loopback += 1
                else:
                    self._frame_stats.in_frames += 1

                cid = CANID.parse(envelope.frame.identifier)
                if cid is not None:  # Ignore non-UAVCAN/CAN frames
                    ufr = UAVCANFrame.parse(envelope.frame)
                    if ufr is not None:  # Ignore non-UAVCAN/CAN frames
                        self._handle_any_frame(timestamp, cid, ufr, loopback=envelope.loopback)
            except Exception as ex:  # pragma: no cover
                self._frame_stats.in_frames_errored += 1
                _logger.exception("%s: Error while processing received %s: %s", self, envelope, ex)

        if self._capture_handlers:  # When capture is enabled, we force loopback for all outgoing frames.
            broadcast = pyuavcan.util.broadcast(self._capture_handlers)
            for timestamp, envelope in frames:
                broadcast(CANCapture(timestamp, envelope.frame, own=envelope.loopback))

    def _handle_any_frame(self, timestamp: Timestamp, can_id: CANID, frame: UAVCANFrame, loopback: bool) -> None:
        if not loopback:
            self._frame_stats.in_frames_uavcan += 1
            if self._handle_received_frame(timestamp, can_id, frame):
                self._frame_stats.in_frames_uavcan_accepted += 1
        else:
            self._handle_loopback_frame(timestamp, can_id, frame)

    def _handle_received_frame(self, timestamp: Timestamp, can_id: CANID, frame: UAVCANFrame) -> bool:
        _logger.debug("%s: Accepted: %s %s %s", self, timestamp, frame, can_id)
        ss = pyuavcan.transport.InputSessionSpecifier(can_id.data_specifier, can_id.source_node_id)
        accepted = False
        dest_nid = can_id.get_destination_node_id()
        if dest_nid is None or dest_nid == self._local_node_id:
            session = self._input_dispatch_table.get(ss)
            if session is not None:
                session._push_frame(timestamp, can_id, frame)  # pylint: disable=protected-access
                accepted = True

            if ss.remote_node_id is not None:
                ss = pyuavcan.transport.InputSessionSpecifier(ss.data_specifier, None)
                session = self._input_dispatch_table.get(ss)
                if session is not None:
                    session._push_frame(timestamp, can_id, frame)  # pylint: disable=protected-access
                    accepted = True

        return accepted

    def _handle_loopback_frame(self, timestamp: Timestamp, can_id: CANID, frame: UAVCANFrame) -> None:
        _logger.debug("%s: Loopback: %s %s %s", self, timestamp, frame, can_id)
        ss = pyuavcan.transport.OutputSessionSpecifier(can_id.data_specifier, can_id.get_destination_node_id())
        try:
            session = self._output_registry[ss]
        except KeyError:
            pass  # Do not log this because packet capture mode generates a lot of unattended loopback frames.
        else:
            session._handle_loopback_frame(timestamp, frame)  # pylint: disable=protected-access

    def _reconfigure_acceptance_filters(self) -> None:
        if not self._capture_handlers:
            subject_ids = set(
                ds.subject_id
                for ds in (x.specifier.data_specifier for x in self._input_dispatch_table.items)
                if isinstance(ds, pyuavcan.transport.MessageDataSpecifier)
            )
            fcs = generate_filter_configurations(subject_ids, self._local_node_id)
            assert len(fcs) > len(subject_ids)
        else:
            fcs = [
                FilterConfiguration.new_promiscuous(FrameFormat.BASE),
                FilterConfiguration.new_promiscuous(FrameFormat.EXTENDED),
            ]

        if self._maybe_media is not None:
            num_filters = self._maybe_media.number_of_acceptance_filters
            fcs = optimize_filter_configurations(fcs, num_filters)
            assert len(fcs) <= num_filters
            if self._last_filter_configuration_set != fcs:
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(
                        "%s: Configuring %d acceptance filters:\n%s", self, num_filters, "\n".join(map(str, fcs))
                    )
                try:
                    self._maybe_media.configure_acceptance_filters(fcs)
                except Exception:  # pragma: no cover
                    self._last_filter_configuration_set = None
                    raise
                else:
                    self._last_filter_configuration_set = fcs

    def _get_repr_fields(self) -> typing.Tuple[typing.List[typing.Any], typing.Dict[str, typing.Any]]:
        return [self._maybe_media], {
            "local_node_id": self.local_node_id,
        }



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import copy
import typing
import asyncio
import logging
import ipaddress
import dataclasses
import pyuavcan
from ._session import UDPInputSession, SelectiveUDPInputSession, PromiscuousUDPInputSession
from ._session import UDPOutputSession
from ._frame import UDPFrame
from ._ip import SocketFactory, Sniffer, LinkLayerCapture, unicast_ip_to_node_id, node_id_to_unicast_ip
from ._socket_reader import SocketReader, SocketReaderStatistics
from ._tracer import UDPTracer, UDPCapture


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class UDPTransportStatistics(pyuavcan.transport.TransportStatistics):
    received_datagrams: typing.Dict[pyuavcan.transport.DataSpecifier, SocketReaderStatistics] = dataclasses.field(
        default_factory=dict
    )
    """
    Basic input session statistics: instances of :class:`SocketReaderStatistics` keyed by data specifier.
    """


class UDPTransport(pyuavcan.transport.Transport):
    """
    The UAVCAN/UDP (IP v4/v6) transport is designed for low-latency, high-throughput, high-reliability
    vehicular networks based on Ethernet.
    Please read the module documentation for details.
    """

    TRANSFER_ID_MODULO = UDPFrame.TRANSFER_ID_MASK + 1

    VALID_MTU_RANGE = 1200, 9000
    """
    The minimum is based on the IPv6 specification, which guarantees that the path MTU is at least 1280 bytes large.
    This value is also acceptable for virtually all IPv4 local or real-time networks.
    Lower MTU values shall not be used because they may lead to multi-frame transfer fragmentation where this is
    not expected by the designer, possibly violating the real-time constraints.

    A conventional Ethernet jumbo frame can carry up to 9 KiB (9216 bytes).
    These are the application-level MTU values, so we take overheads into account.
    """

    VALID_SERVICE_TRANSFER_MULTIPLIER_RANGE = (1, 5)

    def __init__(
        self,
        local_ip_address: typing.Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address],
        local_node_id: typing.Optional[int] = -1,
        *,
        mtu: int = min(VALID_MTU_RANGE),
        service_transfer_multiplier: int = 1,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
        anonymous: bool = False,
    ):
        """
        :param local_ip_address: Specifies which local IP address to use for this transport.
            This setting also implicitly specifies the network interface to use.
            All output sockets will be bound (see ``bind()``) to the specified local address.
            If the specified address is not available locally, the transport will fail with
            :class:`pyuavcan.transport.InvalidMediaConfigurationError`.

            For use on the loopback interface, any IP address from the loopback range can be used;
            for example, ``127.0.0.123``.
            This generally does not work with physical interfaces;
            for example, if a host has one physical interface at ``192.168.1.200``,
            an attempt to run a node at ``192.168.1.201`` will trigger the media configuration error
            because ``bind()`` will fail with ``EADDRNOTAVAIL``.
            One can change the node-ID of a physical transport by altering the network
            interface configuration in the underlying operating system itself.

            Using ``INADDR_ANY`` here (i.e., ``0.0.0.0`` for IPv4) is not expected to work reliably or be portable
            because this configuration is, generally, incompatible with multicast sockets (even in the anonymous mode).
            In order to set up even a listening multicast socket, it is necessary to specify the correct local
            address such that the underlying IP stack is aware of which interface to receive multicast packets from.

            When the anonymous mode is enabled, it is quite possible to snoop on the network even if there is
            another node running locally on the same interface
            (because sockets are initialized with ``SO_REUSEADDR`` and ``SO_REUSEPORT``, when available).

        :param local_node_id: As explained previously, the node-ID is part of the IP address,
            but this parameter allows one to use the UDP transport in anonymous mode or easily build the
            node IP address from a subnet address (like ``127.42.0.0``) and a node-ID.

            - If the value is negative, the node-ID equals the 16 least significant bits of the ``local_ip_address``.
              This is the default behavior.

            - If the value is None, an anonymous instance will be constructed,
              where the transport will reject any attempt to create an output session.
              The transport instance will also report its own :attr:`local_node_id` as None.
              The UAVCAN/UDP transport does not support anonymous transfers by design.

            - If the value is a non-negative integer, the 16 least significant bits of the ``local_ip_address``
              are replaced with this value.

            Examples:

            +-----------------------+-------------------+----------------------------+--------------------------+
            | ``local_ip_address``  | ``local_node_id`` | Local IP address           | Local node-ID            |
            +=======================+===================+============================+==========================+
            | 127.42.1.200          | (default)         | 127.42.1.200               | 456 (from IP address)    |
            +-----------------------+-------------------+----------------------------+--------------------------+
            | 127.42.1.200          | 42                | 127.42.0.42                | 42                       |
            +-----------------------+-------------------+----------------------------+--------------------------+
            | 127.42.0.0            | 42                | 127.42.0.42                | 42                       |
            +-----------------------+-------------------+----------------------------+--------------------------+
            | 127.42.1.200          | None              | 127.42.1.200               | anonymous                |
            +-----------------------+-------------------+----------------------------+--------------------------+

        :param mtu: The application-level MTU for outgoing packets.
            In other words, this is the maximum number of serialized bytes per UAVCAN/UDP frame.
            Transfers where the number of payload bytes does not exceed this value will be single-frame transfers,
            otherwise, multi-frame transfers will be used.
            This setting affects only outgoing frames;
            the MTU of incoming frames is fixed at a sufficiently large value to accept any meaningful UDP frame.

            The default value is the smallest valid value for reasons of compatibility.

        :param service_transfer_multiplier: Deterministic data loss mitigation is disabled by default.
            This parameter specifies the number of times each outgoing service transfer will be repeated.
            This setting does not affect message transfers.

        :param loop: The event loop to use. Defaults to :func:`asyncio.get_event_loop`.

        :param anonymous: DEPRECATED and scheduled for removal; replace with ``local_node_id=None``.
        """
        if anonymous:  # Backward compatibility. Will be removed.
            import warnings

            local_node_id = None
            warnings.warn("Parameter 'anonymous' is deprecated. Use 'local_node_id=None' instead.", DeprecationWarning)

        if not isinstance(local_ip_address, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            local_ip_address = ipaddress.ip_address(local_ip_address)
        assert not isinstance(local_ip_address, str)
        if local_node_id is not None and local_node_id >= 0:
            local_ip_address = node_id_to_unicast_ip(local_ip_address, local_node_id)

        self._sock_factory = SocketFactory.new(local_ip_address)
        self._anonymous = local_node_id is None
        self._mtu = int(mtu)
        self._srv_multiplier = int(service_transfer_multiplier)
        self._loop = loop if loop is not None else asyncio.get_event_loop()

        low, high = self.VALID_SERVICE_TRANSFER_MULTIPLIER_RANGE
        if not (low <= self._srv_multiplier <= high):
            raise ValueError(f"Invalid service transfer multiplier: {self._srv_multiplier}")

        low, high = self.VALID_MTU_RANGE
        if not (low <= self._mtu <= high):
            raise ValueError(f"Invalid MTU: {self._mtu} bytes")

        self._socket_reader_registry: typing.Dict[pyuavcan.transport.DataSpecifier, SocketReader] = {}
        self._input_registry: typing.Dict[pyuavcan.transport.InputSessionSpecifier, UDPInputSession] = {}
        self._output_registry: typing.Dict[pyuavcan.transport.OutputSessionSpecifier, UDPOutputSession] = {}

        self._sniffer: typing.Optional[Sniffer] = None
        self._capture_handlers: typing.List[pyuavcan.transport.CaptureCallback] = []

        self._closed = False
        self._statistics = UDPTransportStatistics()

        assert (local_node_id is None) or (local_node_id < 0) or (self.local_node_id == local_node_id)
        assert (self.local_node_id is None) or (0 <= self.local_node_id <= 0xFFFF)
        _logger.debug("%s: Initialized with local node-ID %s", self, self.local_node_id)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def protocol_parameters(self) -> pyuavcan.transport.ProtocolParameters:
        return pyuavcan.transport.ProtocolParameters(
            transfer_id_modulo=self.TRANSFER_ID_MODULO,
            max_nodes=self._sock_factory.max_nodes,
            mtu=self._mtu,
        )

    @property
    def local_node_id(self) -> typing.Optional[int]:
        addr = self._sock_factory.local_ip_address
        return None if self._anonymous else unicast_ip_to_node_id(addr, addr)

    def close(self) -> None:
        self._closed = True
        for s in (*self.input_sessions, *self.output_sessions):
            try:
                s.close()
            except Exception as ex:  # pragma: no cover
                _logger.exception("%s: Failed to close %r: %s", self, s, ex)
        if self._sniffer is not None:
            self._sniffer.close()
            self._sniffer = None

    def get_input_session(
        self, specifier: pyuavcan.transport.InputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> UDPInputSession:
        self._ensure_not_closed()
        if specifier not in self._input_registry:
            self._setup_input_session(specifier, payload_metadata)
        assert specifier.data_specifier in self._socket_reader_registry
        out = self._input_registry[specifier]
        assert isinstance(out, UDPInputSession)
        assert out.specifier == specifier
        return out

    def get_output_session(
        self, specifier: pyuavcan.transport.OutputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> UDPOutputSession:
        self._ensure_not_closed()
        if specifier not in self._output_registry:
            if self.local_node_id is None:
                # In UAVCAN/UDP, the anonymous mode is somewhat bolted-on.
                # The underlying protocol (IP) does not have the concept of anonymous packet.
                # We add it artificially as an implementation detail of this library.
                raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                    "Cannot create an output session instance because this UAVCAN/UDP transport instance is "
                    "configured in the anonymous mode."
                )

            def finalizer() -> None:
                del self._output_registry[specifier]

            multiplier = (
                self._srv_multiplier
                if isinstance(specifier.data_specifier, pyuavcan.transport.ServiceDataSpecifier)
                else 1
            )
            sock = self._sock_factory.make_output_socket(specifier.remote_node_id, specifier.data_specifier)
            self._output_registry[specifier] = UDPOutputSession(
                specifier=specifier,
                payload_metadata=payload_metadata,
                mtu=self._mtu,
                multiplier=multiplier,
                sock=sock,
                loop=self._loop,
                finalizer=finalizer,
            )

        out = self._output_registry[specifier]
        assert isinstance(out, UDPOutputSession)
        assert out.specifier == specifier
        return out

    def sample_statistics(self) -> UDPTransportStatistics:
        return copy.copy(self._statistics)

    @property
    def input_sessions(self) -> typing.Sequence[UDPInputSession]:
        return list(self._input_registry.values())

    @property
    def output_sessions(self) -> typing.Sequence[UDPOutputSession]:
        return list(self._output_registry.values())

    @property
    def local_ip_address(self) -> typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]:
        return self._sock_factory.local_ip_address

    def begin_capture(self, handler: pyuavcan.transport.CaptureCallback) -> None:
        """
        Reported events are of type :class:`UDPCapture`.

        In order for the network capture to work, the local machine should be connected to a SPAN port of the switch.
        See https://en.wikipedia.org/wiki/Port_mirroring and read the documentation for your networking hardware.
        Additional preconditions must be met depending on the platform:

        - On GNU/Linux, network capture requires that either the process is executed by root,
          or the raw packet capture capability ``CAP_NET_RAW`` is enabled.
          For more info read ``man 7 capabilities`` and consider checking the docs for Wireshark/libpcap.

        - On Windows, Npcap needs to be installed and configured; see https://nmap.org/npcap/.

        Packets that do not originate from the current UAVCAN/UDP subnet (configured on this transport instance)
        are not reported via this interface.
        This restriction is critical because there may be other UAVCAN/UDP networks running on the same physical
        L2 network segregated by different subnets, so that if foreign packets were not dropped,
        conflicts would occur.
        """
        self._ensure_not_closed()
        if self._sniffer is None:
            _logger.debug("%s: Starting UDP/IP packet capture (hope you have permissions)", self)
            self._sniffer = self._sock_factory.make_sniffer(self._process_capture)
        self._capture_handlers.append(handler)

    @property
    def capture_active(self) -> bool:
        return self._sniffer is not None

    @staticmethod
    def make_tracer() -> UDPTracer:
        """
        See :class:`UDPTracer`.
        """
        return UDPTracer()

    async def spoof(self, transfer: pyuavcan.transport.AlienTransfer, monotonic_deadline: float) -> bool:
        """
        Not implemented yet. Always raises :class:`NotImplementedError`.
        When implemented, this method will rely on libpcap to emit spoofed link-layer packets.
        """
        raise NotImplementedError

    def _setup_input_session(
        self, specifier: pyuavcan.transport.InputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> None:
        """
        In order to set up a new input session, we have to link together a lot of objects. Tricky.
        Also, the setup and teardown actions shall be atomic. Hence the separate method.
        """
        assert specifier not in self._input_registry
        try:
            if specifier.data_specifier not in self._socket_reader_registry:
                _logger.debug(
                    "%r: Setting up new socket reader for %s. Existing entries at the moment: %s",
                    self,
                    specifier.data_specifier,
                    self._socket_reader_registry,
                )
                self._socket_reader_registry[specifier.data_specifier] = SocketReader(
                    sock=self._sock_factory.make_input_socket(specifier.data_specifier),
                    local_ip_address=self._sock_factory.local_ip_address,
                    anonymous=self._anonymous,
                    statistics=self._statistics.received_datagrams.setdefault(
                        specifier.data_specifier, SocketReaderStatistics()
                    ),
                    loop=self.loop,
                )

            cls: typing.Union[typing.Type[PromiscuousUDPInputSession], typing.Type[SelectiveUDPInputSession]] = (
                PromiscuousUDPInputSession if specifier.is_promiscuous else SelectiveUDPInputSession
            )

            session = cls(
                specifier=specifier,
                payload_metadata=payload_metadata,
                loop=self.loop,
                finalizer=lambda: self._teardown_input_session(specifier),
            )
            self._socket_reader_registry[specifier.data_specifier].add_listener(
                specifier.remote_node_id, session._process_frame  # pylint: disable=protected-access
            )
        except Exception:
            self._teardown_input_session(specifier)  # Rollback to ensure atomicity.
            raise

        self._input_registry[specifier] = session

    def _teardown_input_session(self, specifier: pyuavcan.transport.InputSessionSpecifier) -> None:
        """
        The finalizer may be invoked at any point during the setup process,
        so it must be able to deconstruct the pipeline even if it is not fully set up.
        This is why we have these try-except everywhere. Who knew that atomic transactions can be so messy?
        """
        # Unregister the session first.
        try:
            del self._input_registry[specifier]
        except LookupError:
            pass
        # Remove the session from the list of socket reader listeners.
        try:
            demux = self._socket_reader_registry[specifier.data_specifier]
        except LookupError:
            pass  # The reader has not been set up yet, nothing to do.
        else:
            try:
                demux.remove_listener(specifier.remote_node_id)
            except LookupError:
                pass
            # Destroy the reader if there are no listeners left.
            if not demux.has_listeners:
                try:
                    _logger.debug("%r: Destroying %r for %s", self, demux, specifier.data_specifier)
                    demux.close()
                finally:
                    del self._socket_reader_registry[specifier.data_specifier]

    def _process_capture(self, capture: LinkLayerCapture) -> None:
        """This handler may be invoked from a different thread (the capture thread)."""
        pyuavcan.util.broadcast(self._capture_handlers)(UDPCapture(capture.timestamp, capture.packet))

    def _ensure_not_closed(self) -> None:
        if self._closed:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

    def _get_repr_fields(self) -> typing.Tuple[typing.List[typing.Any], typing.Dict[str, typing.Any]]:
        return [repr(str(self.local_ip_address))], {
            "local_node_id": self.local_node_id,
            "service_transfer_multiplier": self._srv_multiplier,
            "mtu": self._mtu,
        }



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
from ..crc import CRC32C

TransferCRC = CRC32C



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import time
import typing
import decimal
import datetime


_AnyScalar = typing.Union[float, int, decimal.Decimal]

_DECIMAL_NANO = decimal.Decimal("1e-9")


class Timestamp:
    """
    Timestamps are hashable and immutable.
    Timestamps can be compared for exact equality; relational comparison operators are not defined.

    A timestamp instance always contains a pair of time samples:
    the *system time*, also known as "wall time" or local civil time,
    and the monotonic time, which is used only for time interval measurement.
    """

    def __init__(self, system_ns: int, monotonic_ns: int) -> None:
        """
        Manual construction is rarely needed, except when implementing network drivers.
        See the static factory methods.

        :param system_ns:       Belongs to the domain of :func:`time.time_ns`. Units are nanoseconds.
        :param monotonic_ns:    Belongs to the domain of :func:`time.monotonic_ns`. Units are nanoseconds.
        """
        self._system_ns = int(system_ns)
        self._monotonic_ns = int(monotonic_ns)

        if self._system_ns < 0 or self._monotonic_ns < 0:
            raise ValueError(f"Neither of the timestamp samples can be negative; found this: {self!r}")

    @staticmethod
    def from_seconds(system: _AnyScalar, monotonic: _AnyScalar) -> Timestamp:
        """
        Both inputs are in seconds (not nanoseconds) of any numerical type.
        """
        return Timestamp(system_ns=Timestamp._second_to_ns(system), monotonic_ns=Timestamp._second_to_ns(monotonic))

    @staticmethod
    def now() -> Timestamp:
        """
        Constructs a new timestamp instance populated with current time.

        .. important:: Clocks are sampled non-atomically! Monotonic sampled first.
        """
        return Timestamp(monotonic_ns=time.monotonic_ns(), system_ns=time.time_ns())

    @staticmethod
    def combine_oldest(*arguments: Timestamp) -> Timestamp:
        """
        Picks lowest time values from the provided set of timestamps and constructs a new instance from those.

        This can be useful for transfer reception logic where the oldest frame timestamp is used as the
        transfer timestamp for multi-frame transfers to reduce possible timestamping error variation
        introduced in the media layer.

        >>> Timestamp.combine_oldest(
        ...     Timestamp(12345, 45600),
        ...     Timestamp(12300, 45699),
        ...     Timestamp(12399, 45678),
        ... )
        Timestamp(system_ns=12300, monotonic_ns=45600)
        """
        return Timestamp(
            system_ns=min(x.system_ns for x in arguments), monotonic_ns=min(x.monotonic_ns for x in arguments)
        )

    @property
    def system(self) -> decimal.Decimal:
        """System time in seconds."""
        return self._ns_to_second(self._system_ns)

    @property
    def monotonic(self) -> decimal.Decimal:
        """Monotonic time in seconds."""
        return self._ns_to_second(self._monotonic_ns)

    @property
    def system_ns(self) -> int:
        return self._system_ns

    @property
    def monotonic_ns(self) -> int:
        return self._monotonic_ns

    @staticmethod
    def _second_to_ns(x: _AnyScalar) -> int:
        return int(decimal.Decimal(x) / _DECIMAL_NANO)

    @staticmethod
    def _ns_to_second(x: int) -> decimal.Decimal:
        return decimal.Decimal(x) * _DECIMAL_NANO

    def __eq__(self, other: typing.Any) -> bool:
        """
        Performs an exact comparison of the timestamp components with nanosecond resolution.
        """
        if isinstance(other, Timestamp):
            return self._system_ns == other._system_ns and self._monotonic_ns == other._monotonic_ns
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._system_ns + self._monotonic_ns)

    def __str__(self) -> str:
        dt = datetime.datetime.fromtimestamp(float(self.system))  # Precision loss is OK - system time is imprecise
        iso = dt.isoformat(timespec="microseconds")
        return f"{iso}/{self.monotonic:.6f}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}(system_ns={self._system_ns}, monotonic_ns={self._monotonic_ns})"


def _unittest_timestamp() -> None:
    from pytest import raises
    from decimal import Decimal

    Timestamp(0, 0)

    with raises(ValueError):
        Timestamp(-1, 0)

    with raises(ValueError):
        Timestamp(0, -1)

    ts = Timestamp.from_seconds(Decimal("5.123456789"), Decimal("123.456789"))
    assert ts.system_ns == 5123456789
    assert ts.monotonic_ns == 123456789000
    assert ts.system == Decimal("5.123456789")
    assert ts.monotonic == Decimal("123.456789")
    assert hash(ts) == hash(Timestamp(5123456789, 123456789000))
    assert hash(ts) != hash(Timestamp(123, 456))
    assert ts == Timestamp(5123456789, 123456789000)
    assert ts != Timestamp(123, 123456789000)
    assert ts != Timestamp(5123456789, 456)
    assert ts != "Hello"
    assert Timestamp.combine_oldest(Timestamp(123, 123456789000), Timestamp(5123456789, 456), ts) == Timestamp(123, 456)
    print(ts)



# Copyright (C) 2021  UAVCAN Consortium  <uavcan.org>
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
Implementation of the UAVCAN register interface as defined in the UAVCAN Specification
(section 5.3 *Application-layer functions*).
"""

from uavcan.primitive import Empty_1_0 as Empty
from uavcan.primitive import String_1_0 as String
from uavcan.primitive import Unstructured_1_0 as Unstructured
from uavcan.primitive.array import Bit_1_0 as Bit
from uavcan.primitive.array import Integer64_1_0 as Integer64
from uavcan.primitive.array import Integer32_1_0 as Integer32
from uavcan.primitive.array import Integer16_1_0 as Integer16
from uavcan.primitive.array import Integer8_1_0 as Integer8
from uavcan.primitive.array import Natural64_1_0 as Natural64
from uavcan.primitive.array import Natural32_1_0 as Natural32
from uavcan.primitive.array import Natural16_1_0 as Natural16
from uavcan.primitive.array import Natural8_1_0 as Natural8
from uavcan.primitive.array import Real64_1_0 as Real64
from uavcan.primitive.array import Real32_1_0 as Real32
from uavcan.primitive.array import Real16_1_0 as Real16

from ._value import Value as Value
from ._value import ValueProxy as ValueProxy
from ._value import RelaxedValue as RelaxedValue
from ._value import ValueConversionError as ValueConversionError

from . import backend as backend

from ._registry import Registry as Registry
from ._registry import ValueProxyWithFlags as ValueProxyWithFlags
from ._registry import MissingRegisterError as MissingRegisterError


def get_environment_variable_name(register_name: str) -> str:
    """
    Convert the name of the register to the name of the environment variable that assigns it.

    >>> get_environment_variable_name("m.motor.inductance_dq")
    'M__MOTOR__INDUCTANCE_DQ'
    """
    return register_name.upper().replace(".", "__")



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._media import Media as Media

from ._frame import FrameFormat as FrameFormat
from ._frame import DataFrame as DataFrame
from ._frame import Envelope as Envelope

from ._filter import FilterConfiguration as FilterConfiguration
from ._filter import optimize_filter_configurations as optimize_filter_configurations



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._input import UDPInputSession as UDPInputSession
from ._input import PromiscuousUDPInputSession as PromiscuousUDPInputSession
from ._input import SelectiveUDPInputSession as SelectiveUDPInputSession

from ._input import UDPInputSessionStatistics as UDPInputSessionStatistics
from ._input import PromiscuousUDPInputSessionStatistics as PromiscuousUDPInputSessionStatistics
from ._input import SelectiveUDPInputSessionStatistics as SelectiveUDPInputSessionStatistics

from ._output import UDPOutputSession as UDPOutputSession
from ._output import UDPFeedback as UDPFeedback
